@@ -21,6 +21,7 @@
 #pragma optimize("gty", on)
 #endif
 
+#include "ppsspp_config.h"
 #ifdef _WIN32
 #include "Common/CommonWindows.h"
 #include <WindowsX.h>
@@ -30,16 +31,34 @@
 
 #include <algorithm>
 #include <memory>
+#ifndef NO_ARMIPS
+#include <string_view>
+#endif
+#include <unordered_map>
 
-#include "util/text/utf8.h"
 #include "zlib.h"
+
 #include "Common/CommonTypes.h"
-#include "Common/FileUtil.h"
+#include "Common/Data/Encoding/Utf8.h"
+#include "Common/Log.h"
+#include "Common/File/FileUtil.h"
 #include "Common/StringUtils.h"
 #include "Core/MemMap.h"
 #include "Core/Debugger/SymbolMap.h"
 
+#ifndef NO_ARMIPS
 #include "ext/armips/Core/Assembler.h"
+#else
+struct Identifier {
+	explicit Identifier() {}
+	explicit Identifier(const std::string &s) {}
+};
+
+struct LabelDefinition {
+	Identifier name;
+	int64_t value;
+};
+#endif
 
 SymbolMap *g_symbolMap;
 
@@ -62,15 +81,17 @@ void SymbolMap::Clear() {
 	activeNeedUpdate_ = false;
 }
 
-bool SymbolMap::LoadSymbolMap(const char *filename) {
+bool SymbolMap::LoadSymbolMap(const Path &filename) {
 	Clear();  // let's not recurse the lock
 
 	std::lock_guard<std::recursive_mutex> guard(lock_);
 
+	// TODO(scoped): Use gzdopen instead.
+
 #if defined(_WIN32) && defined(UNICODE)
-	gzFile f = gzopen_w(ConvertUTF8ToWString(filename).c_str(), "r");
+	gzFile f = gzopen_w(filename.ToWString().c_str(), "r");
 #else
-	gzFile f = gzopen(filename, "r");
+	gzFile f = gzopen(filename.c_str(), "r");
 #endif
 
 	if (f == Z_NULL)
@@ -120,9 +141,9 @@ bool SymbolMap::LoadSymbolMap(const char *filename) {
 
 		if (!started) continue;
 
-		u32 address = -1, size, vaddress = -1;
+		u32 address = -1, size = 0, vaddress = -1;
 		int moduleIndex = 0;
-		int typeInt;
+		int typeInt = ST_NONE;
 		SymbolType type;
 		char name[128] = {0};
 
@@ -138,7 +159,9 @@ bool SymbolMap::LoadSymbolMap(const char *filename) {
 			continue;
 		}
 
-		sscanf(line, "%08x %08x %x %i %127c", &address, &size, &vaddress, &typeInt, name);
+		int matched = sscanf(line, "%08x %08x %x %i %127c", &address, &size, &vaddress, &typeInt, name);
+		if (matched < 1)
+			continue;
 		type = (SymbolType) typeInt;
 		if (!hasModules) {
 			if (!Memory::IsValidAddress(vaddress)) {
@@ -183,22 +206,23 @@ bool SymbolMap::LoadSymbolMap(const char *filename) {
 	return started;
 }
 
-void SymbolMap::SaveSymbolMap(const char *filename) const {
+bool SymbolMap::SaveSymbolMap(const Path &filename) const {
 	std::lock_guard<std::recursive_mutex> guard(lock_);
 
 	// Don't bother writing a blank file.
 	if (!File::Exists(filename) && functions.empty() && data.empty()) {
-		return;
+		return true;
 	}
 
+	// TODO(scoped): Use gzdopen
 #if defined(_WIN32) && defined(UNICODE)
-	gzFile f = gzopen_w(ConvertUTF8ToWString(filename).c_str(), "w9");
+	gzFile f = gzopen_w(filename.ToWString().c_str(), "w9");
 #else
-	gzFile f = gzopen(filename, "w9");
+	gzFile f = gzopen(filename.c_str(), "w9");
 #endif
 
 	if (f == Z_NULL)
-		return;
+		return false;
 
 	gzprintf(f, ".text\n");
 
@@ -217,9 +241,10 @@ void SymbolMap::SaveSymbolMap(const char *filename) const {
 		gzprintf(f, "%08x %08x %x %i %s\n", e.start, e.size, e.module, ST_DATA, GetLabelNameRel(e.start, e.module));
 	}
 	gzclose(f);
+	return true;
 }
 
-bool SymbolMap::LoadNocashSym(const char *filename) {
+bool SymbolMap::LoadNocashSym(const Path &filename) {
 	std::lock_guard<std::recursive_mutex> guard(lock_);
 	FILE *f = File::OpenCFile(filename, "r");
 	if (!f)
@@ -277,7 +302,7 @@ bool SymbolMap::LoadNocashSym(const char *filename) {
 	return true;
 }
 
-void SymbolMap::SaveNocashSym(const char *filename) const {
+void SymbolMap::SaveNocashSym(const Path &filename) const {
 	std::lock_guard<std::recursive_mutex> guard(lock_);
 
 	// Don't bother writing a blank file.
@@ -432,7 +457,7 @@ void SymbolMap::AddModule(const char *name, u32 address, u32 size) {
 			// Just reactivate that one.
 			it->start = address;
 			it->size = size;
-			activeModuleEnds.insert(std::make_pair(it->start + it->size, *it));
+			activeModuleEnds.emplace(it->start + it->size, *it);
 			activeNeedUpdate_ = true;
 			return;
 		}
@@ -445,7 +470,7 @@ void SymbolMap::AddModule(const char *name, u32 address, u32 size) {
 	mod.index = (int)modules.size() + 1;
 
 	modules.push_back(mod);
-	activeModuleEnds.insert(std::make_pair(mod.start + mod.size, mod));
+	activeModuleEnds.emplace(mod.start + mod.size, mod);
 	activeNeedUpdate_ = true;
 }
 
@@ -552,7 +577,7 @@ void SymbolMap::AddFunction(const char* name, u32 address, u32 size, int moduleI
 		auto active = activeFunctions.find(address);
 		if (active != activeFunctions.end() && active->second.module == moduleIndex) {
 			activeFunctions.erase(active);
-			activeFunctions.insert(std::make_pair(address, existing->second));
+			activeFunctions.emplace(address, existing->second);
 		}
 	} else {
 		FunctionEntry func;
@@ -563,7 +588,7 @@ void SymbolMap::AddFunction(const char* name, u32 address, u32 size, int moduleI
 		functions[symbolKey] = func;
 
 		if (IsModuleActive(moduleIndex)) {
-			activeFunctions.insert(std::make_pair(address, func));
+			activeFunctions.emplace(address, func);
 		}
 	}
 
@@ -613,8 +638,31 @@ u32 SymbolMap::FindPossibleFunctionAtAfter(u32 address) {
 }
 
 u32 SymbolMap::GetFunctionSize(u32 startAddress) {
-	if (activeNeedUpdate_)
-		UpdateActiveSymbols();
+	if (activeNeedUpdate_) {
+		std::lock_guard<std::recursive_mutex> guard(lock_);
+
+		// This is common, from the jit.  Direct lookup is faster than updating active symbols.
+		auto mod = activeModuleEnds.lower_bound(startAddress);
+		std::pair<int, u32> funcKey;
+		if (mod == activeModuleEnds.end()) {
+			// Could still be mod 0, backwards compatibility.
+			if (!sawUnknownModule)
+				return INVALID_ADDRESS;
+			funcKey.first = 0;
+			funcKey.second = startAddress;
+		} else {
+			if (mod->second.start > startAddress)
+				return INVALID_ADDRESS;
+			funcKey.first = mod->second.index;
+			funcKey.second = startAddress - mod->second.start;
+		}
+
+		auto func = functions.find(funcKey);
+		if (func == functions.end())
+			return INVALID_ADDRESS;
+
+		return func->second.size;
+	}
 
 	std::lock_guard<std::recursive_mutex> guard(lock_);
 	auto it = activeFunctions.find(startAddress);
@@ -678,7 +726,7 @@ void SymbolMap::UpdateActiveSymbols() {
 		return;
 	}
 
-	std::map<int, u32> activeModuleIndexes;
+	std::unordered_map<int, u32> activeModuleIndexes;
 	for (auto it = activeModuleEnds.begin(), end = activeModuleEnds.end(); it != end; ++it) {
 		activeModuleIndexes[it->second.index] = it->second.start;
 	}
@@ -686,27 +734,27 @@ void SymbolMap::UpdateActiveSymbols() {
 	for (auto it = functions.begin(), end = functions.end(); it != end; ++it) {
 		const auto mod = activeModuleIndexes.find(it->second.module);
 		if (it->second.module == 0) {
-			activeFunctions.insert(std::make_pair(it->second.start, it->second));
+			activeFunctions.emplace(it->second.start, it->second);
 		} else if (mod != activeModuleIndexes.end()) {
-			activeFunctions.insert(std::make_pair(mod->second + it->second.start, it->second));
+			activeFunctions.emplace(mod->second + it->second.start, it->second);
 		}
 	}
 
 	for (auto it = labels.begin(), end = labels.end(); it != end; ++it) {
 		const auto mod = activeModuleIndexes.find(it->second.module);
 		if (it->second.module == 0) {
-			activeLabels.insert(std::make_pair(it->second.addr, it->second));
+			activeLabels.emplace(it->second.addr, it->second);
 		} else if (mod != activeModuleIndexes.end()) {
-			activeLabels.insert(std::make_pair(mod->second + it->second.addr, it->second));
+			activeLabels.emplace(mod->second + it->second.addr, it->second);
 		}
 	}
 
 	for (auto it = data.begin(), end = data.end(); it != end; ++it) {
 		const auto mod = activeModuleIndexes.find(it->second.module);
 		if (it->second.module == 0) {
-			activeData.insert(std::make_pair(it->second.start, it->second));
+			activeData.emplace(it->second.start, it->second);
 		} else if (mod != activeModuleIndexes.end()) {
-			activeData.insert(std::make_pair(mod->second + it->second.start, it->second));
+			activeData.emplace(mod->second + it->second.start, it->second);
 		}
 	}
 
@@ -727,7 +775,7 @@ bool SymbolMap::SetFunctionSize(u32 startAddress, u32 newSize) {
 		if (func != functions.end()) {
 			func->second.size = newSize;
 			activeFunctions.erase(funcInfo);
-			activeFunctions.insert(std::make_pair(startAddress, func->second));
+			activeFunctions.emplace(startAddress, func->second);
 		}
 	}
 
@@ -799,7 +847,7 @@ void SymbolMap::AddLabel(const char* name, u32 address, int moduleIndex) {
 			auto active = activeLabels.find(address);
 			if (active != activeLabels.end() && active->second.module == moduleIndex) {
 				activeLabels.erase(active);
-				activeLabels.insert(std::make_pair(address, label));
+				activeLabels.emplace(address, label);
 			}
 		}
 	} else {
@@ -810,7 +858,7 @@ void SymbolMap::AddLabel(const char* name, u32 address, int moduleIndex) {
 
 		labels[symbolKey] = label;
 		if (IsModuleActive(moduleIndex)) {
-			activeLabels.insert(std::make_pair(address, label));
+			activeLabels.emplace(address, label);
 		}
 	}
 }
@@ -834,7 +882,7 @@ void SymbolMap::SetLabelName(const char* name, u32 address) {
 			auto active = activeLabels.find(address);
 			if (active != activeLabels.end() && active->second.module == label->second.module) {
 				activeLabels.erase(active);
-				activeLabels.insert(std::make_pair(address, label->second));
+				activeLabels.emplace(address, label->second);
 			}
 		}
 	}
@@ -917,7 +965,7 @@ void SymbolMap::AddData(u32 address, u32 size, DataType type, int moduleIndex) {
 		auto active = activeData.find(address);
 		if (active != activeData.end() && active->second.module == moduleIndex) {
 			activeData.erase(active);
-			activeData.insert(std::make_pair(address, existing->second));
+			activeData.emplace(address, existing->second);
 		}
 	} else {
 		DataEntry entry;
@@ -928,7 +976,7 @@ void SymbolMap::AddData(u32 address, u32 size, DataType type, int moduleIndex) {
 
 		data[symbolKey] = entry;
 		if (IsModuleActive(moduleIndex)) {
-			activeData.insert(std::make_pair(address, entry));
+			activeData.emplace(address, entry);
 		}
 	}
 }
@@ -1007,7 +1055,9 @@ void SymbolMap::GetLabels(std::vector<LabelDefinition> &dest) {
 	for (auto it = activeLabels.begin(); it != activeLabels.end(); it++) {
 		LabelDefinition entry;
 		entry.value = it->first;
-		entry.name = ConvertUTF8ToWString(it->second.name);
+		std::string name = it->second.name;
+		std::transform(name.begin(), name.end(), name.begin(), ::tolower);
+		entry.name = Identifier(name);
 		dest.push_back(entry);
 	}
 }
@@ -1043,7 +1093,6 @@ void SymbolMap::FillSymbolListBox(HWND listbox,SymbolType symType) {
 			SendMessage(listbox, LB_INITSTORAGE, (WPARAM)activeFunctions.size(), (LPARAM)activeFunctions.size() * 30);
 
 			for (auto it = activeFunctions.begin(), end = activeFunctions.end(); it != end; ++it) {
-				const FunctionEntry& entry = it->second;
 				const char* name = GetLabelName(it->first);
 				if (name != NULL)
 					wsprintf(temp, L"%S", name);
@@ -1067,7 +1116,6 @@ void SymbolMap::FillSymbolListBox(HWND listbox,SymbolType symType) {
 			}
 
 			for (auto it = activeData.begin(), end = activeData.end(); it != end; ++it) {
-				const DataEntry& entry = it->second;
 				const char* name = GetLabelName(it->first);
 
 				if (name != NULL)
