@@ -15,24 +15,25 @@
 // Official git repository and contact information can be found at
 // https://github.com/hrydgard/ppsspp and http://www.ppsspp.org/.
 
-#include "i18n/i18n.h"
-#include "math/math_util.h"
+#include "ppsspp_config.h"
 #include <algorithm>
 
+#include "Common/Data/Text/I18n.h"
+#include "Common/Math/math_util.h"
+#include "Common/Data/Encoding/Utf8.h"
+#include "Common/Serialize/SerializeFuncs.h"
+#include "Common/System/Request.h"
+#include "Common/Serialize/Serializer.h"
+
 #include "Core/Dialog/PSPOskDialog.h"
 #include "Core/Util/PPGeDraw.h"
 #include "Core/HLE/sceCtrl.h"
-#include "Core/HLE/sceDisplay.h"
 #include "Core/HLE/sceUtility.h"
+#include "Core/HW/Display.h"
 #include "Core/Config.h"
 #include "Core/Reporting.h"
-#include "Common/ChunkFile.h"
 #include "GPU/GPUState.h"
 
-#if defined(USING_WIN_UI)
-#include "base/NativeApp.h"
-#endif
-
 #ifndef _WIN32
 #include <ctype.h>
 #include <math.h>
@@ -45,8 +46,8 @@ const static int OSK_SHUTDOWN_DELAY_US = 40000;
 
 static std::map<std::string, std::pair<std::string, int>> languageMapping;
 
-const int numKeyCols[OSK_KEYBOARD_COUNT] = {12, 12, 13, 13, 12, 12, 12, 12, 12};
-const int numKeyRows[OSK_KEYBOARD_COUNT] = {4, 4, 5, 5, 5, 4, 4, 4, 4};
+const uint8_t numKeyCols[OSK_KEYBOARD_COUNT] = {12, 12, 13, 13, 12, 12, 12, 12, 12};
+const uint8_t numKeyRows[OSK_KEYBOARD_COUNT] = {4, 4, 6, 6, 5, 4, 4, 4, 4};
 
 // Japanese (Kana) diacritics
 static const wchar_t diacritics[2][103] =
@@ -62,89 +63,102 @@ static const wchar_t kor_cons[] = L"ㄱㄲㄴㄷㄸㄹㅁㅂㅃㅅㅆㅇㅈㅉ
 static const wchar_t kor_vowel[] = L"ㅏㅐㅑㅒㅓㅔㅕㅖㅗ   ㅛㅜ   ㅠㅡ ㅣ";
 
 // Korean (Hangul) vowel Combination key
-const int kor_vowelCom[] = {0,8,9,1,8,10,20,8,11,4,13,14,5,13,15,20,13,16,20,18,19};
+const uint8_t kor_vowelCom[] = {0,8,9,1,8,10,20,8,11,4,13,14,5,13,15,20,13,16,20,18,19};
 
 // Korean (Hangul) last consonant(diacritics)
 static const wchar_t kor_lcons[] = L"ㄱㄲㄳㄴㄵㄶㄷㄹㄺㄻㄼㄽㄾㄿㅀㅁㅂㅄㅅㅆㅇㅈㅊㅋㅌㅍㅎ";
 
 // Korean (Hangul) last consonant Combination key
-const int kor_lconsCom[] = {18,0,2,21,3,4,26,3,5,0,7,8,15,7,9,16,7,10,18,7,11,24,7,12,25,7,13,26,7,14,18,16,17};
+const uint8_t kor_lconsCom[] = {18,0,2,21,3,4,26,3,5,0,7,8,15,7,9,16,7,10,18,7,11,24,7,12,25,7,13,26,7,14,18,16,17};
 
 // Korean (Hangul) last consonant Separation key
-const int kor_lconsSpr[] = {2,1,9,4,4,12,5,4,18,8,8,0,9,8,6,10,8,7,11,8,9,12,8,16,13,8,17,14,8,18,17,17,9};
+const uint8_t kor_lconsSpr[] = {2,1,9,4,4,12,5,4,18,8,8,0,9,8,6,10,8,7,11,8,9,12,8,16,13,8,17,14,8,18,17,17,9};
 
-static const wchar_t oskKeys[OSK_KEYBOARD_COUNT][5][14] =
+static const char16_t oskKeys[OSK_KEYBOARD_COUNT][6][14] =
 {
 	{
 		// Latin Lowercase
-		{L"1234567890-+"},
-		{L"qwertyuiop[]"},
-		{L"asdfghjkl;@~"},
-		{L"zxcvbnm,./?\\"},
+		{u"1234567890-+"},
+		{u"qwertyuiop[]"},
+		{u"asdfghjkl;@~"},
+		{u"zxcvbnm,./?\\"},
 	},
 	{
 		// Latin Uppercase
-		{L"!@#$%^&*()_+"},
-		{L"QWERTYUIOP{}"},
-		{L"ASDFGHJKL:\"`"},
-		{L"ZXCVBNM<>/?|"},
+		{u"!@#$%^&*()_+"},
+		{u"QWERTYUIOP{}"},
+		{u"ASDFGHJKL:\"`"},
+		{u"ZXCVBNM<>/?|"},
 	},
 	{
 		// Hiragana
-		{L"あかさたなはまやらわぁゃっ"},
-		{L"いきしちにひみ　り　ぃ　　"},
-		{L"うくすつぬふむゆるをぅゅ゛"},
-		{L"えけせてねへめ　れ　ぇ　゜"},
-		{L"おこそとのほもよろんぉょー"},
+		{u"あかさたなはまやらわぁゃっ"},
+		{u"いきしちにひみ　り　ぃ　　"},
+		{u"うくすつぬふむゆるをぅゅ゛"},
+		{u"えけせてねへめ　れ　ぇ　゜"},
+		{u"おこそとのほもよろんぉょー"},
+		{u"・。、「」『』〜     "},
 	},
 	{
 		// Katakana
-		{L"アカサタナハマヤラワァャッ"},
-		{L"イキシチニヒミ　リ　ィ　　"},
-		{L"ウクスツヌフムユルヲゥュ゛"},
-		{L"エケセテネヘメ　レ　ェ　゜"},
-		{L"オコソトノホモヨロンォョー"},
+		{u"アカサタナハマヤラワァャッ"},
+		{u"イキシチニヒミ　リ　ィ　　"},
+		{u"ウクスツヌフムユルヲゥュ゛"},
+		{u"エケセテネヘメ　レ　ェ　゜"},
+		{u"オコソトノホモヨロンォョー"},
+		{u"・。、「」『』〜     "},
 	},
 	{
 		// Korean(Hangul)
-		{L"1234567890-+"},
-		{L"ㅃㅉㄸㄲㅆ!@#$%^&"},
-		{L"ㅂㅈㄷㄱㅅㅛㅕㅑㅐㅔ[]"},
-		{L"ㅁㄴㅇㄹㅎㅗㅓㅏㅣ;@~"},
-		{L"ㅋㅌㅊㅍㅠㅜㅡ<>/?|"},
+		{u"1234567890-+"},
+		{u"ㅃㅉㄸㄲㅆ!@#$%^&"},
+		{u"ㅂㅈㄷㄱㅅㅛㅕㅑㅐㅔ[]"},
+		{u"ㅁㄴㅇㄹㅎㅗㅓㅏㅣ;@~"},
+		{u"ㅋㅌㅊㅍㅠㅜㅡ<>/?|"},
 	},
 	{
 		// Russian Lowercase
-		{L"1234567890-+"},
-		{L"йцукенгшщзхъ"},
-		{L"фывапролджэё"},
-		{L"ячсмитьбю/?|"},
+		{u"1234567890-+"},
+		{u"йцукенгшщзхъ"},
+		{u"фывапролджэё"},
+		{u"ячсмитьбю/?|"},
 	},
 	{
 		// Russian Uppercase
-		{L"!@#$%^&*()_+"},
-		{L"ЙЦУКЕНГШЩЗХЪ"},
-		{L"ФЫВАПРОЛДЖЭЁ"},
-		{L"ЯЧСМИТЬБЮ/?|"},
+		{u"!@#$%^&*()_+"},
+		{u"ЙЦУКЕНГШЩЗХЪ"},
+		{u"ФЫВАПРОЛДЖЭЁ"},
+		{u"ЯЧСМИТЬБЮ/?|"},
 	},
 	{
 		// Latin Full-width Lowercase
-		{ L"１２３４５６７８９０－＋" },
-		{ L"ｑｗｅｒｔｙｕｉｏｐ［］" },
-		{ L"ａｓｄｆｇｈｊｋｌ；＠～" },
-		{ L"ｚｘｃｖｂｎｍ，．／？￥￥" },
+		{ u"１２３４５６７８９０－＋" },
+		{ u"ｑｗｅｒｔｙｕｉｏｐ［］" },
+		{ u"ａｓｄｆｇｈｊｋｌ；＠～" },
+		{ u"ｚｘｃｖｂｎｍ，．／？￥￥" },
 	},
 	{
 		// Latin Full-width Uppercase
-		{ L"！＠＃＄％＾＆＊（）＿＋" },
-		{ L"ＱＷＥＲＴＹＵＩＯＰ｛｝" },
-		{ L"ＡＳＤＦＧＨＪＫＬ：￥”‘" },
-		{ L"ＺＸＣＶＢＮＭ＜＞／？｜" },
+		{ u"！＠＃＄％＾＆＊（）＿＋" },
+		{ u"ＱＷＥＲＴＹＵＩＯＰ｛｝" },
+		{ u"ＡＳＤＦＧＨＪＫＬ：￥”‘" },
+		{ u"ＺＸＣＶＢＮＭ＜＞／？｜" },
 	},
 };
 
 // This isn't a complete representation of these flags, it just helps ensure we show the right keyboards.
-int allowedInputFlagsMap[OSK_KEYBOARD_COUNT] = {
+static const int allowedInputFlagsMap[OSK_KEYBOARD_COUNT] = {
+	PSP_UTILITY_OSK_INPUTTYPE_LATIN_LOWERCASE | PSP_UTILITY_OSK_INPUTTYPE_LATIN_UPPERCASE | PSP_UTILITY_OSK_INPUTTYPE_LATIN_SYMBOL | PSP_UTILITY_OSK_INPUTTYPE_LATIN_DIGIT,
+	PSP_UTILITY_OSK_INPUTTYPE_LATIN_LOWERCASE | PSP_UTILITY_OSK_INPUTTYPE_LATIN_UPPERCASE | PSP_UTILITY_OSK_INPUTTYPE_LATIN_SYMBOL,
+	PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_HIRAGANA,
+	PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_KATAKANA | PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_HALF_KATAKANA,
+	PSP_UTILITY_OSK_INPUTTYPE_KOREAN,
+	PSP_UTILITY_OSK_INPUTTYPE_RUSSIAN_LOWERCASE | PSP_UTILITY_OSK_INPUTTYPE_RUSSIAN_UPPERCASE,
+	PSP_UTILITY_OSK_INPUTTYPE_RUSSIAN_LOWERCASE | PSP_UTILITY_OSK_INPUTTYPE_RUSSIAN_UPPERCASE,
+	PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_LOWERCASE | PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_UPPERCASE | PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_SYMBOL | PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_DIGIT,
+	PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_LOWERCASE | PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_UPPERCASE | PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_SYMBOL,
+};
+static const int defaultInputFlagsMap[OSK_KEYBOARD_COUNT] = {
 	PSP_UTILITY_OSK_INPUTTYPE_LATIN_LOWERCASE | PSP_UTILITY_OSK_INPUTTYPE_LATIN_SYMBOL | PSP_UTILITY_OSK_INPUTTYPE_LATIN_DIGIT,
 	PSP_UTILITY_OSK_INPUTTYPE_LATIN_UPPERCASE | PSP_UTILITY_OSK_INPUTTYPE_LATIN_SYMBOL,
 	PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_HIRAGANA,
@@ -156,7 +170,7 @@ int allowedInputFlagsMap[OSK_KEYBOARD_COUNT] = {
 	PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_UPPERCASE | PSP_UTILITY_OSK_INPUTTYPE_JAPANESE_SYMBOL,
 };
 
-PSPOskDialog::PSPOskDialog() : PSPDialog() {
+PSPOskDialog::PSPOskDialog(UtilityDialogType type) : PSPDialog(type) {
 	// This can break all kinds of stuff, changing the decimal point in sprintf for example.
 	// Not sure what the intended effect is so commented out for now.
 	// setlocale(LC_ALL, "");
@@ -169,7 +183,7 @@ void PSPOskDialog::ConvertUCS2ToUTF8(std::string& _string, const PSPPointer<u16_
 {
 	if (!em_address.IsValid())
 	{
-		_string = "";
+		_string.clear();
 		return;
 	}
 
@@ -196,17 +210,17 @@ void PSPOskDialog::ConvertUCS2ToUTF8(std::string& _string, const PSPPointer<u16_
 	_string = stringBuffer;
 }
 
-void GetWideStringFromPSPPointer(std::wstring& _string, const PSPPointer<u16_le>& em_address)
+void GetWideStringFromPSPPointer(std::u16string& _string, const PSPPointer<u16_le>& em_address)
 {
 	if (!em_address.IsValid())
 	{
-		_string = L"";
+		_string.clear();
 		return;
 	}
 
 	const size_t maxLength = 2047;
-	wchar_t stringBuffer[maxLength + 1];
-	wchar_t *string = stringBuffer;
+	char16_t stringBuffer[maxLength + 1];
+	char16_t *string = stringBuffer;
 
 	u16_le *input = &em_address[0];
 	int c;
@@ -216,7 +230,7 @@ void GetWideStringFromPSPPointer(std::wstring& _string, const PSPPointer<u16_le>
 	_string = stringBuffer;
 }
 
-void PSPOskDialog::ConvertUCS2ToUTF8(std::string& _string, const wchar_t *input)
+void PSPOskDialog::ConvertUCS2ToUTF8(std::string& _string, const char16_t *input)
 {
 	char stringBuffer[2048];
 	char *string = stringBuffer;
@@ -246,11 +260,17 @@ static void FindValidKeyboard(s32 inputType, int direction, OskKeyboardLanguage
 	if (inputType == 0) {
 		return;
 	}
+	// We use direction = 0 for default, but we actually move "forward".
+	const int *matchMap = allowedInputFlagsMap;
+	if (direction == 0) {
+		direction = 1;
+		matchMap = defaultInputFlagsMap;
+	}
 
 	// TODO: Limit by allowed keyboards properly... this is just an approximation.
 	int tries = OSK_LANGUAGE_COUNT * 2;
-	while (!(inputType & allowedInputFlagsMap[disp]) && tries > 0) {
-		if ((--tries % 1) == 0) {
+	while (!(inputType & matchMap[disp]) && tries > 0) {
+		if ((--tries % 2) == 0) {
 			lang = (OskKeyboardLanguage)((OSK_LANGUAGE_COUNT + lang + direction) % OSK_LANGUAGE_COUNT);
 			disp = OskKeyboardCases[lang][LOWERCASE];
 		} else {
@@ -309,7 +329,7 @@ int PSPOskDialog::Init(u32 oskPtr) {
 	selectedChar = 0;
 	currentKeyboardLanguage = OSK_LANGUAGE_ENGLISH;
 	currentKeyboard = OSK_KEYBOARD_LATIN_LOWERCASE;
-	FindValidKeyboard(oskParams->fields[0].inputtype, 1, currentKeyboardLanguage, currentKeyboard);
+	FindValidKeyboard(oskParams->fields[0].inputtype, 0, currentKeyboardLanguage, currentKeyboard);
 
 	ConvertUCS2ToUTF8(oskDesc, oskParams->fields[0].desc);
 	ConvertUCS2ToUTF8(oskIntext, oskParams->fields[0].intext);
@@ -317,7 +337,7 @@ int PSPOskDialog::Init(u32 oskPtr) {
 
 	i_level = 0;
 
-	inputChars = L"";
+	inputChars.clear();
 
 	if (oskParams->fields[0].intext.IsValid()) {
 		auto src = oskParams->fields[0].intext;
@@ -326,25 +346,29 @@ int PSPOskDialog::Init(u32 oskPtr) {
 			inputChars += c;
 	}
 
-	languageMapping = GetLangValuesMapping();
+	languageMapping = g_Config.GetLangValuesMapping();
 
 	// Eat any keys pressed before the dialog inited.
 	UpdateButtons();
+	InitCommon();
+
+	std::lock_guard<std::mutex> guard(nativeMutex_);
+	nativeStatus_ = PSPOskNativeStatus::IDLE;
 
 	StartFade(true);
 	return 0;
 }
 
-std::wstring PSPOskDialog::CombinationKorean(bool isInput)
+std::u16string PSPOskDialog::CombinationKorean(bool isInput)
 {
-	std::wstring string;
+	std::u16string string;
 
 	isCombinated = true;
 
 	int selectedRow = selectedChar / numKeyCols[currentKeyboard];
 	int selectedCol = selectedChar % numKeyCols[currentKeyboard];
 
-	if(inputChars.size() == 0) {
+	if (inputChars.size() == 0) {
 		wchar_t sw = oskKeys[currentKeyboard][selectedRow][selectedCol];
 
 		if (inputChars.size() < FieldMaxLength()) {
@@ -577,9 +601,9 @@ std::wstring PSPOskDialog::CombinationKorean(bool isInput)
 	return string;
 }
 
-std::wstring PSPOskDialog::CombinationString(bool isInput)
+std::u16string PSPOskDialog::CombinationString(bool isInput)
 {
-	std::wstring string;
+	std::u16string string;
 
 	isCombinated = false;
 
@@ -658,7 +682,7 @@ std::wstring PSPOskDialog::CombinationString(bool isInput)
 				string += inputChars[i];
 			}
 
-			if (string.size() <= FieldMaxLength())
+			if (string.size() < FieldMaxLength())
 			{
 				string += oskKeys[currentKeyboard][selectedRow][selectedCol];
 			}
@@ -748,10 +772,15 @@ u32 PSPOskDialog::FieldMaxLength()
 
 void PSPOskDialog::RenderKeyboard()
 {
+	// Sanity check that a valid keyboard is selected.
+	if ((int)currentKeyboard < 0 || (int)currentKeyboard >= OSK_KEYBOARD_COUNT) {
+		return;
+	}
+
 	int selectedRow = selectedChar / numKeyCols[currentKeyboard];
 	int selectedCol = selectedChar % numKeyCols[currentKeyboard];
 
-	wchar_t temp[2];
+	char16_t temp[2];
 	temp[1] = '\0';
 
 	std::string buffer;
@@ -765,10 +794,16 @@ void PSPOskDialog::RenderKeyboard()
 	float previewLeftSide = (480.0f - (12.0f * drawLimit)) / 2.0f;
 	float title = (480.0f - (0.5f * drawLimit)) / 2.0f;
 
+	PPGeStyle descStyle = FadedStyle(PPGeAlign::BOX_CENTER, 0.5f);
+	PPGeDrawText(oskDesc.c_str(), title, 20, descStyle);
 
-	PPGeDrawText(oskDesc.c_str(), title , 20, PPGE_ALIGN_CENTER, 0.5f, CalcFadedColor(0xFFFFFFFF));
+	PPGeStyle textStyle = FadedStyle(PPGeAlign::BOX_HCENTER, 0.5f);
 
-	std::wstring result;
+	PPGeStyle keyStyle = FadedStyle(PPGeAlign::BOX_HCENTER, 0.6f);
+	PPGeStyle selectedKeyStyle = FadedStyle(PPGeAlign::BOX_HCENTER, 0.6f);
+	selectedKeyStyle.color = CalcFadedColor(0xFF3060FF);
+
+	std::u16string result;
 
 	result = CombinationString(false);
 
@@ -776,12 +811,11 @@ void PSPOskDialog::RenderKeyboard()
 	drawIndex = result.size() == limit + 1 ? drawIndex - 1 : drawIndex;  // When the length reached limit, the last character don't fade in and out.
 	for (u32 i = 0; i < drawLimit; ++i, ++drawIndex)
 	{
-		u32 color = CalcFadedColor(0xFFFFFFFF);
 		if (drawIndex + 1 < result.size())
 		{
 			temp[0] = result[drawIndex];
 			ConvertUCS2ToUTF8(buffer, temp);
-			PPGeDrawText(buffer.c_str(), previewLeftSide + (i * characterWidth), 40.0f, PPGE_ALIGN_HCENTER, 0.5f, color);
+			PPGeDrawText(buffer.c_str(), previewLeftSide + (i * characterWidth), 40.0f, textStyle);
 		}
 		else
 		{
@@ -794,25 +828,25 @@ void PSPOskDialog::RenderKeyboard()
 					float animStep = (float)(__DisplayGetNumVblanks() % 40) / 20.0f;
 					// Fade in and out the next character so they know it's not part of the string yet.
 					u32 alpha = (0.5f - (cosf(animStep * M_PI) / 2.0f)) * 128 + 127;
-					color = CalcFadedColor((alpha << 24) | 0xFFFFFF);
+					PPGeStyle animStyle = textStyle;
+					animStyle.color = CalcFadedColor((alpha << 24) | 0x00FFFFFF);
 
 					ConvertUCS2ToUTF8(buffer, temp);
 
-					PPGeDrawText(buffer.c_str(), previewLeftSide + (i * characterWidth), 40.0f, PPGE_ALIGN_HCENTER, 0.5f, color);
+					PPGeDrawText(buffer.c_str(), previewLeftSide + (i * characterWidth), 40.0f, animStyle);
 
 					// Also draw the underline for the same reason.
-					color = CalcFadedColor(0xFFFFFFFF);
-					PPGeDrawText("_", previewLeftSide + (i * characterWidth), 40.0f, PPGE_ALIGN_HCENTER, 0.5f, color);
+					PPGeDrawText("_", previewLeftSide + (i * characterWidth), 40.0f, textStyle);
 				}
 				else
 				{
 					ConvertUCS2ToUTF8(buffer, temp);
-					PPGeDrawText(buffer.c_str(), previewLeftSide + (i * characterWidth), 40.0f, PPGE_ALIGN_HCENTER, 0.5f, color);
+					PPGeDrawText(buffer.c_str(), previewLeftSide + (i * characterWidth), 40.0f, textStyle);
 				}
 			}
 			else
 			{
-				PPGeDrawText("_", previewLeftSide + (i * characterWidth), 40.0f, PPGE_ALIGN_HCENTER, 0.5f, color);
+				PPGeDrawText("_", previewLeftSide + (i * characterWidth), 40.0f, textStyle);
 			}
 		}
 	}
@@ -821,22 +855,20 @@ void PSPOskDialog::RenderKeyboard()
 	{
 		for (int col = 0; col < numKeyCols[currentKeyboard]; ++col)
 		{
-			u32 color = CalcFadedColor(0xFFFFFFFF);
-			if (selectedRow == row && col == selectedCol)
-				color = CalcFadedColor(0xFF3060FF);
-
 			temp[0] = oskKeys[currentKeyboard][row][col];
 
 			ConvertUCS2ToUTF8(buffer, temp);
-			PPGeDrawText(buffer.c_str(), keyboardLeftSide + (25.0f * col) + characterWidth / 2.0, 70.0f + (25.0f * row), PPGE_ALIGN_HCENTER, 0.6f, color);
 
-			if (selectedRow == row && col == selectedCol)
-				PPGeDrawText("_", keyboardLeftSide + (25.0f * col) + characterWidth / 2.0, 70.0f + (25.0f * row), PPGE_ALIGN_HCENTER, 0.6f, CalcFadedColor(0xFFFFFFFF));
+			if (selectedRow == row && col == selectedCol) {
+				PPGeDrawText(buffer.c_str(), keyboardLeftSide + (25.0f * col) + characterWidth / 2.0, 70.0f + (25.0f * row), selectedKeyStyle);
+				PPGeDrawText("_", keyboardLeftSide + (25.0f * col) + characterWidth / 2.0, 70.0f + (25.0f * row), keyStyle);
+			} else {
+				PPGeDrawText(buffer.c_str(), keyboardLeftSide + (25.0f * col) + characterWidth / 2.0, 70.0f + (25.0f * row), keyStyle);
+			}
 		}
 	}
 }
 
-#if defined(USING_WIN_UI)
 // TODO: Why does this have a 2 button press lag/delay when
 // re-opening the dialog box? I don't get it.
 int PSPOskDialog::NativeKeyboard() {
@@ -844,23 +876,63 @@ int PSPOskDialog::NativeKeyboard() {
 		return SCE_ERROR_UTILITY_INVALID_STATUS;
 	}
 
-	std::wstring titleText;
-	GetWideStringFromPSPPointer(titleText, oskParams->fields[0].desc);
+#if defined(USING_WIN_UI) || defined(USING_QT_UI) || PPSSPP_PLATFORM(ANDROID)
+	bool beginInputBox = false;
+	if (nativeStatus_ == PSPOskNativeStatus::IDLE) {
+		std::lock_guard<std::mutex> guard(nativeMutex_);
+		if (nativeStatus_ == PSPOskNativeStatus::IDLE) {
+			nativeStatus_ = PSPOskNativeStatus::WAITING;
+			beginInputBox = true;
+		}
+	}
 
-	std::wstring defaultText;
-	GetWideStringFromPSPPointer(defaultText, oskParams->fields[0].intext);
+	if (beginInputBox) {
+		std::u16string titleText;
+		GetWideStringFromPSPPointer(titleText, oskParams->fields[0].desc);
 
-	if (defaultText.empty())
-		defaultText.assign(L"VALUE");
+		std::u16string defaultText;
+		GetWideStringFromPSPPointer(defaultText, oskParams->fields[0].intext);
+
+		if (defaultText.empty())
+			defaultText.assign(u"VALUE");
+
+		// There's already ConvertUCS2ToUTF8 in this file. Should we use that instead of the global ones?
+		System_InputBoxGetString(::ConvertUCS2ToUTF8(titleText), ::ConvertUCS2ToUTF8(defaultText),
+			[&](const std::string &value, int) {
+				// Success callback
+				std::lock_guard<std::mutex> guard(nativeMutex_);
+				if (nativeStatus_ != PSPOskNativeStatus::WAITING) {
+					return;
+				}
+				nativeValue_ = value;
+				nativeStatus_ = PSPOskNativeStatus::SUCCESS;
+			},
+			[&]() {
+				// Failure callback
+				std::lock_guard<std::mutex> guard(nativeMutex_);
+				if (nativeStatus_ != PSPOskNativeStatus::WAITING) {
+					return;
+				}
+				nativeValue_ = "";
+				nativeStatus_ = PSPOskNativeStatus::FAILURE;
+			}
+		);
+	} else if (nativeStatus_ == PSPOskNativeStatus::SUCCESS) {
+		inputChars = ConvertUTF8ToUCS2(nativeValue_);
+		nativeValue_.clear();
 
-	if (System_InputBoxGetWString(titleText.c_str(), defaultText, inputChars)) {
 		u32 maxLength = FieldMaxLength();
 		if (inputChars.length() > maxLength) {
 			ERROR_LOG(SCEUTILITY, "NativeKeyboard: input text too long(%d characters/glyphs max), truncating to game-requested length.", maxLength);
 			inputChars.erase(maxLength, std::string::npos);
 		}
+		ChangeStatus(SCE_UTILITY_STATUS_FINISHED, 0);
+		nativeStatus_ = PSPOskNativeStatus::DONE;
+	} else if (nativeStatus_ == PSPOskNativeStatus::FAILURE) {
+		ChangeStatus(SCE_UTILITY_STATUS_FINISHED, 0);
+		nativeStatus_ = PSPOskNativeStatus::DONE;
 	}
-	ChangeStatus(SCE_UTILITY_STATUS_FINISHED, 0);
+#endif
 	
 	u16_le *outText = oskParams->fields[0].outtext;
 
@@ -870,7 +942,7 @@ int PSPOskDialog::NativeKeyboard() {
 	// Only write the bytes of the output and the null terminator, don't write the rest.
 	for (size_t i = 0; i < end; ++i) {
 		u16 value = 0;
-		if (i < FieldMaxLength())
+		if (i < FieldMaxLength() && i < inputChars.size())
 			value = inputChars[i];
 		outText[i] = value;
 	}
@@ -880,15 +952,15 @@ int PSPOskDialog::NativeKeyboard() {
 
 	return 0;
 }
-#endif
 
 int PSPOskDialog::Update(int animSpeed) {
 	if (GetStatus() != SCE_UTILITY_STATUS_RUNNING) {
 		return SCE_ERROR_UTILITY_INVALID_STATUS;
 	}
 
-	int cancelButton = g_Config.iButtonPreference == PSP_SYSTEMPARAM_BUTTON_CROSS ? CTRL_CIRCLE : CTRL_CROSS;
-	int confirmButton = cancelButton == CTRL_CROSS ? CTRL_CIRCLE : CTRL_CROSS;
+	int cancelButton = GetCancelButton();
+	int confirmButton = GetConfirmButton();
+
 	static int cancelBtnFramesHeld = 0;
 	static int confirmBtnFramesHeld = 0;
 	static int leftBtnFramesHeld = 0;
@@ -899,16 +971,17 @@ int PSPOskDialog::Update(int animSpeed) {
 	const int framesHeldRepeatRate = 5;
 
 	UpdateButtons();
+	UpdateCommon();
 	int selectedRow = selectedChar / numKeyCols[currentKeyboard];
 	int selectedExtra = selectedChar % numKeyCols[currentKeyboard];
 
-	// TODO: Add your platforms here when you have a NativeKeyboard func.
-
-#if defined(USING_WIN_UI)
+#if defined(USING_WIN_UI) || defined(USING_QT_UI) || PPSSPP_PLATFORM(ANDROID) || PPSSPP_PLATFORM(SWITCH)
 	// Windows: Fall back to the OSK/continue normally if we're in fullscreen.
 	// The dialog box doesn't work right if in fullscreen.
-	if(g_Config.bBypassOSKWithKeyboard && !g_Config.bFullScreen)
-		return NativeKeyboard();
+	if (System_GetPropertyBool(SYSPROP_HAS_KEYBOARD)) {
+		if (g_Config.bBypassOSKWithKeyboard && !g_Config.UseFullScreen())
+			return NativeKeyboard();
+	}
 #endif
 
 	UpdateFade(animSpeed);
@@ -917,24 +990,27 @@ int PSPOskDialog::Update(int animSpeed) {
 	PPGeDrawRect(0, 0, 480, 272, CalcFadedColor(0x63636363));
 	RenderKeyboard();
 
-	I18NCategory *di = GetI18NCategory("Dialog");
+	auto di = GetI18NCategory(I18NCat::DIALOG);
 
-	PPGeDrawImage(I_SQUARE, 365, 222, 16, 16, 0, CalcFadedColor(0xFFFFFFFF));
-	PPGeDrawText(di->T("Space"), 390, 222, PPGE_ALIGN_LEFT, 0.5f, CalcFadedColor(0xFFFFFFFF));
+	PPGeStyle actionStyle = FadedStyle(PPGeAlign::BOX_LEFT, 0.5f);
+	PPGeStyle guideStyle = FadedStyle(PPGeAlign::BOX_LEFT, 0.6f);
 
-	if (g_Config.iButtonPreference != PSP_SYSTEMPARAM_BUTTON_CIRCLE) {
-		PPGeDrawImage(I_CROSS, 45, 222, 16, 16, 0, CalcFadedColor(0xFFFFFFFF));
-		PPGeDrawImage(I_CIRCLE, 45, 247, 16, 16, 0, CalcFadedColor(0xFFFFFFFF));
+	PPGeDrawImage(ImageID("I_SQUARE"), 365, 222, 16, 16, guideStyle);
+	PPGeDrawText(di->T("Space"), 390, 222, actionStyle);
+
+	if (GetConfirmButton() != CTRL_CIRCLE) {
+		PPGeDrawImage(ImageID("I_CROSS"), 45, 222, 16, 16, guideStyle);
+		PPGeDrawImage(ImageID("I_CIRCLE"), 45, 247, 16, 16, guideStyle);
 	} else {
-		PPGeDrawImage(I_CIRCLE, 45, 222, 16, 16, 0, CalcFadedColor(0xFFFFFFFF));
-		PPGeDrawImage(I_CROSS, 45, 247, 16, 16, 0, CalcFadedColor(0xFFFFFFFF));
+		PPGeDrawImage(ImageID("I_CIRCLE"), 45, 222, 16, 16, guideStyle);
+		PPGeDrawImage(ImageID("I_CROSS"), 45, 247, 16, 16, guideStyle);
 	}
 
-	PPGeDrawText(di->T("Select"), 75, 222, PPGE_ALIGN_LEFT, 0.5f, CalcFadedColor(0xFFFFFFFF));
-	PPGeDrawText(di->T("Delete"), 75, 247, PPGE_ALIGN_LEFT, 0.5f, CalcFadedColor(0xFFFFFFFF));
+	PPGeDrawText(di->T("Select"), 75, 222, actionStyle);
+	PPGeDrawText(di->T("Delete"), 75, 247, actionStyle);
 
-	PPGeDrawText("Start", 135, 220, PPGE_ALIGN_LEFT, 0.6f, CalcFadedColor(0xFFFFFFFF));
-	PPGeDrawText(di->T("Finish"), 185, 222, PPGE_ALIGN_LEFT, 0.5f, CalcFadedColor(0xFFFFFFFF));
+	PPGeDrawText("Start", 135, 220, guideStyle);
+	PPGeDrawText(di->T("Finish"), 185, 222, actionStyle);
 
 	auto lookupLangName = [&](int direction) {
 		// First, find the valid one...
@@ -958,20 +1034,20 @@ int PSPOskDialog::Update(int animSpeed) {
 	};
 
 	if (OskKeyboardNames[currentKeyboardLanguage] != "ko_KR" && IsKeyboardShiftValid(oskParams->fields[0].inputtype, currentKeyboardLanguage, currentKeyboard)) {
-		PPGeDrawText("Select", 135, 245, PPGE_ALIGN_LEFT, 0.6f, CalcFadedColor(0xFFFFFFFF));
-		PPGeDrawText(di->T("Shift"), 185, 247, PPGE_ALIGN_LEFT, 0.5f, CalcFadedColor(0xFFFFFFFF));
+		PPGeDrawText("Select", 135, 245, guideStyle);
+		PPGeDrawText(di->T("Shift"), 185, 247, actionStyle);
 	}
 
 	const char *prevLang = lookupLangName(-1);
 	if (prevLang) {
-		PPGeDrawText("L", 235, 220, PPGE_ALIGN_LEFT, 0.6f, CalcFadedColor(0xFFFFFFFF));
-		PPGeDrawText(prevLang, 255, 222, PPGE_ALIGN_LEFT, 0.5f, CalcFadedColor(0xFFFFFFFF));
+		PPGeDrawText("L", 235, 220, guideStyle);
+		PPGeDrawText(prevLang, 255, 222, actionStyle);
 	}
 
 	const char *nextLang = lookupLangName(1);
 	if (nextLang) {
-		PPGeDrawText("R", 235, 245, PPGE_ALIGN_LEFT, 0.6f, CalcFadedColor(0xFFFFFFFF));
-		PPGeDrawText(nextLang, 255, 247, PPGE_ALIGN_LEFT, 0.5f, CalcFadedColor(0xFFFFFFFF));
+		PPGeDrawText("R", 235, 245, guideStyle);
+		PPGeDrawText(nextLang, 255, 247, actionStyle);
 	}
 
 	if (IsButtonPressed(CTRL_UP) || IsButtonHeld(CTRL_UP, upBtnFramesHeld, framesHeldThreshold, framesHeldRepeatRate)) {
@@ -1055,9 +1131,9 @@ int PSPOskDialog::Update(int animSpeed) {
 	} else if (IsButtonPressed(CTRL_SQUARE) && inputChars.size() < FieldMaxLength()) {
 		// Use a regular space if the current keyboard isn't Japanese nor full-width English
 		if (currentKeyboardLanguage != OSK_LANGUAGE_JAPANESE && currentKeyboardLanguage != OSK_LANGUAGE_ENGLISH_FW)
-			inputChars += L" ";
+			inputChars += u" ";
 		else
-			inputChars += L"　";
+			inputChars += u"　";
 	}
 
 	EndDraw();
@@ -1070,7 +1146,7 @@ int PSPOskDialog::Update(int animSpeed) {
 	for (size_t i = 0; i < end; ++i)
 	{
 		u16 value = 0;
-		if (i < inputChars.size())
+		if (i < FieldMaxLength() && i < inputChars.size())
 			value = inputChars[i];
 		outText[i] = value;
 	}
@@ -1089,6 +1165,7 @@ int PSPOskDialog::Shutdown(bool force)
 	if (!force) {
 		ChangeStatusShutdown(OSK_SHUTDOWN_DELAY_US);
 	}
+	nativeStatus_ = PSPOskNativeStatus::IDLE;
 
 	return 0;
 }
@@ -1097,16 +1174,25 @@ void PSPOskDialog::DoState(PointerWrap &p)
 {
 	PSPDialog::DoState(p);
 
-	auto s = p.Section("PSPOskDialog", 1);
+	auto s = p.Section("PSPOskDialog", 1, 2);
 	if (!s)
 		return;
 
-	p.Do(oskParams);
-	p.Do(oskDesc);
-	p.Do(oskIntext);
-	p.Do(oskOuttext);
-	p.Do(selectedChar);
-	p.Do(inputChars);
+	// TODO: Should we save currentKeyboard/currentKeyboardLanguage?
+
+	Do(p, oskParams);
+	Do(p, oskDesc);
+	Do(p, oskIntext);
+	Do(p, oskOuttext);
+	Do(p, selectedChar);
+	if (s >= 2) {
+		Do(p, inputChars);
+	} else {
+		// Discard the wstring.
+		std::wstring wstr;
+		Do(p, wstr);
+	}
+	// Don't need to save state native status or value.
 }
 
 pspUtilityDialogCommon *PSPOskDialog::GetCommonParam()
