@@ -20,9 +20,9 @@
 #include <ctype.h>
 #include <algorithm>
 
-#include "Common/Common.h"
 #include "Common/CommonTypes.h"
-#include "Common/ChunkFile.h"
+#include "Common/Serialize/Serializer.h"
+#include "Common/Serialize/SerializeFuncs.h"
 #include "Core/FileSystems/ISOFileSystem.h"
 #include "Core/HLE/sceKernel.h"
 #include "Core/MemMap.h"
@@ -62,13 +62,27 @@ bool parseLBN(std::string filename, u32 *sectorStart, u32 *readSize) {
 
 #pragma pack(push)
 #pragma pack(1)
+struct u32_le_be_pair {
+	u8 valueLE[4];
+	u8 valueBE[4];
+	operator u32() const {
+		return valueLE[0] + (valueLE[1] << 8) + (valueLE[2] << 16) + (valueLE[3] << 24);
+	}
+};
+
+struct u16_le_be_pair {
+	u8 valueLE[2];
+	u8 valueBE[2];
+	operator u16() const {
+		return valueLE[0] + (valueLE[1] << 8);
+	}
+};
+
 struct DirectoryEntry {
 	u8 size;
 	u8 sectorsInExtendedRecord;
-	u32_le firstDataSectorLE;       // LBA
-	u32_be firstDataSectorBE;
-	u32_le dataLengthLE;            // Size
-	u32_be dataLengthBE;
+	u32_le_be_pair firstDataSector;       // LBA
+	u32_le_be_pair dataLength;            // Size
 	u8 years;
 	u8 month;
 	u8 day;
@@ -79,38 +93,9 @@ struct DirectoryEntry {
 	u8 flags;                       // 2 = directory
 	u8 fileUnitSize;
 	u8 interleaveGap;
-	u16_le volSeqNumberLE;
-	u16_be volSeqNumberBE;
+	u16_le_be_pair volSeqNumber;
 	u8 identifierLength;            //identifier comes right after
 	u8 firstIdChar;
-
-#if COMMON_LITTLE_ENDIAN
-	u32 firstDataSector() const
-	{
-		return firstDataSectorLE;
-	}
-	u32 dataLength() const
-	{
-		return dataLengthLE;
-	}
-	u32 volSeqNumber() const
-	{
-		return volSeqNumberLE;
-	}
-#else
-	u32 firstDataSector() const
-	{
-		return firstDataSectorBE;
-	}
-	u32 dataLength() const
-	{
-		return dataLengthBE;
-	}
-	u32 volSeqNumber() const
-	{
-		return volSeqNumberBE;
-	}
-#endif
 };
 
 struct DirectorySector {
@@ -125,25 +110,16 @@ struct VolDescriptor {
 	char sysid[32];
 	char volid[32];
 	char zeros[8];
-	u32_le numSectorsLE;
-	u32_be numSectoreBE;
+	u32_le_be_pair numSectors;
 	char morezeros[32];
-	u16_le volSetSizeLE;
-	u16_be volSetSizeBE;
-	u16_le volSeqNumLE;
-	u16_be volSeqNumBE;
-	u16_le sectorSizeLE;
-	u16_be sectorSizeBE;
-	u32_le pathTableLengthLE;
-	u32_be pathTableLengthBE;
-	u16_le firstLETableSectorLE;
-	u16_be firstLETableSectorBE;
-	u16_le secondLETableSectorLE;
-	u16_be secondLETableSectorBE;
-	u16_le firstBETableSectorLE;
-	u16_be firstBETableSectorBE;
-	u16_le secondBETableSectorLE;
-	u16_be secondBETableSectorBE;
+	u16_le_be_pair volSetSize;
+	u16_le_be_pair volSeqNum;
+	u16_le_be_pair sectorSize;
+	u32_le_be_pair pathTableLength;
+	u16_le_be_pair firstLETableSector;
+	u16_le_be_pair secondLETableSector;
+	u16_le_be_pair firstBETableSector;
+	u16_le_be_pair secondBETableSector;
 	DirectoryEntry root;
 	char volumeSetIdentifier[128];
 	char publisherIdentifier[128];
@@ -169,9 +145,10 @@ ISOFileSystem::ISOFileSystem(IHandleAllocator *_hAlloc, BlockDevice *_blockDevic
 	hAlloc = _hAlloc;
 
 	VolDescriptor desc;
-	blockDevice->ReadBlock(16, (u8*)&desc);
+	if (!blockDevice->ReadBlock(16, (u8*)&desc))
+		blockDevice->NotifyReadError();
 
-	entireISO.name = "";
+	entireISO.name.clear();
 	entireISO.isDirectory = false;
 	entireISO.startingPosition = 0;
 	entireISO.size = _blockDevice->GetNumBlocks();
@@ -191,8 +168,8 @@ ISOFileSystem::ISOFileSystem(IHandleAllocator *_hAlloc, BlockDevice *_blockDevic
 		return;
 	}
 
-	treeroot->startsector = desc.root.firstDataSector();
-	treeroot->dirsize = desc.root.dataLength();
+	treeroot->startsector = desc.root.firstDataSector;
+	treeroot->dirsize = desc.root.dataLength;
 }
 
 ISOFileSystem::~ISOFileSystem() {
@@ -200,12 +177,20 @@ ISOFileSystem::~ISOFileSystem() {
 	delete treeroot;
 }
 
+std::string ISOFileSystem::TreeEntry::BuildPath() {
+	if (parent) {
+		return parent->BuildPath() + "/" + name;
+	} else {
+		return name;
+	}
+}
+
 void ISOFileSystem::ReadDirectory(TreeEntry *root) {
 	for (u32 secnum = root->startsector, endsector = root->startsector + (root->dirsize + 2047) / 2048; secnum < endsector; ++secnum) {
 		u8 theSector[2048];
 		if (!blockDevice->ReadBlock(secnum, theSector)) {
 			blockDevice->NotifyReadError();
-			ERROR_LOG(FILESYS, "Error reading block for directory %s - skipping", root->name.c_str());
+			ERROR_LOG(FILESYS, "Error reading block for directory '%s' in sector %d - skipping", root->name.c_str(), secnum);
 			root->valid = true;  // Prevents re-reading
 			return;
 		}
@@ -243,16 +228,21 @@ void ISOFileSystem::ReadDirectory(TreeEntry *root) {
 				relative = false;
 			}
 
-			entry->size = dir.dataLength();
-			entry->startingPosition = dir.firstDataSector() * 2048;
+			entry->size = dir.dataLength;
+			entry->startingPosition = dir.firstDataSector * 2048;
 			entry->isDirectory = !isFile;
 			entry->flags = dir.flags;
 			entry->parent = root;
-			entry->startsector = dir.firstDataSector();
-			entry->dirsize = dir.dataLength();
+			entry->startsector = dir.firstDataSector;
+			entry->dirsize = dir.dataLength;
 			entry->valid = isFile;  // Can pre-mark as valid if file, as we don't recurse into those.
-			// Let's not excessively spam the log - I commented this line out.
-			//DEBUG_LOG(FILESYS, "%s: %s %08x %08x %i", entry->isDirectory?"D":"F", entry->name.c_str(), dir.firstDataSectorLE, entry->startingPosition, entry->startingPosition);
+			VERBOSE_LOG(FILESYS, "%s: %s %08x %08x %d", entry->isDirectory ? "D" : "F", entry->name.c_str(), (u32)dir.firstDataSector, entry->startingPosition, entry->startingPosition);
+
+			// Round down to avoid any false reports.
+			if (isFile && dir.firstDataSector + (dir.dataLength / 2048) > blockDevice->GetNumBlocks()) {
+				blockDevice->NotifyReadError();
+				ERROR_LOG(FILESYS, "File '%s' starts or ends outside ISO. firstDataSector: %d len: %d", entry->BuildPath().c_str(), (int)dir.firstDataSector, (int)dir.dataLength);
+			}
 
 			if (entry->isDirectory && !relative) {
 				if (entry->startsector == root->startsector) {
@@ -323,32 +313,37 @@ ISOFileSystem::TreeEntry *ISOFileSystem::GetFromPath(const std::string &path, bo
 				return entry;
 		} else {
 			if (catchError)
-				ERROR_LOG(FILESYS,"File %s not found", path.c_str());
+				ERROR_LOG(FILESYS, "File '%s' not found", path.c_str());
 
 			return 0;
 		}
 	}
 }
 
-u32 ISOFileSystem::OpenFile(std::string filename, FileAccess access, const char *devicename) {
+int ISOFileSystem::OpenFile(std::string filename, FileAccess access, const char *devicename) {
 	OpenFileEntry entry;
 	entry.isRawSector = false;
 	entry.isBlockSectorMode = false;
 
+	if (access & FILEACCESS_WRITE) {
+		ERROR_LOG(FILESYS, "Can't open file '%s' with write access on an ISO partition", filename.c_str());
+		return SCE_KERNEL_ERROR_ERRNO_INVALID_FLAG;
+	}
+
 	if (filename.compare(0, 8, "/sce_lbn") == 0) {
 		// Raw sector read.
 		u32 sectorStart = 0xFFFFFFFF, readSize = 0xFFFFFFFF;
 		parseLBN(filename, &sectorStart, &readSize);
 		if (sectorStart > blockDevice->GetNumBlocks()) {
-			WARN_LOG(FILESYS, "Unable to open raw sector, out of range: %s, sector %08x, max %08x", filename.c_str(), sectorStart, blockDevice->GetNumBlocks());
-			return 0;
+			WARN_LOG(FILESYS, "Unable to open raw sector, out of range: '%s', sector %08x, max %08x", filename.c_str(), sectorStart, blockDevice->GetNumBlocks());
+			return SCE_KERNEL_ERROR_ERRNO_FILE_NOT_FOUND;
 		}
 		else if (sectorStart == blockDevice->GetNumBlocks())
 		{
 			ERROR_LOG(FILESYS, "Should not be able to open the block after the last on disc! %08x", sectorStart);
 		}
 
-		DEBUG_LOG(FILESYS, "Got a raw sector open: %s, sector %08x, size %08x", filename.c_str(), sectorStart, readSize);
+		DEBUG_LOG(FILESYS, "Got a raw sector open: '%s', sector %08x, size %08x", filename.c_str(), sectorStart, readSize);
 		u32 newHandle = hAlloc->GetNewHandle();
 		entry.seekPos = 0;
 		entry.file = 0;
@@ -357,22 +352,17 @@ u32 ISOFileSystem::OpenFile(std::string filename, FileAccess access, const char
 		entry.openSize = readSize;
 		// when open as "umd1:/sce_lbn0x0_size0x6B49D200", that mean open umd1 as a block device.
 		// the param in sceIoLseek and sceIoRead is lba mode. we must mark it.
-		if (strncmp(devicename, "umd0:", 5)==0 || strncmp(devicename, "umd1:", 5)==0)
+		if (strncmp(devicename, "umd0:", 5) == 0 || strncmp(devicename, "umd1:", 5) == 0)
 			entry.isBlockSectorMode = true;
 
 		entries[newHandle] = entry;
 		return newHandle;
 	}
 
-	if (access & FILEACCESS_WRITE) {
-		ERROR_LOG(FILESYS, "Can't open file %s with write access on an ISO partition", filename.c_str());
-		return 0;
-	}
-
-	// May return entireISO for "umd0:"
-	entry.file = GetFromPath(filename);
-	if (!entry.file){
-		return 0;
+	// May return entireISO for "umd0:".
+	entry.file = GetFromPath(filename, false);
+	if (!entry.file) {
+		return SCE_KERNEL_ERROR_ERRNO_FILE_NOT_FOUND;
 	}
 
 	if (entry.file == &entireISO)
@@ -419,13 +409,13 @@ int ISOFileSystem::Ioctl(u32 handle, u32 cmd, u32 indataPtr, u32 inlen, u32 outd
 			return SCE_KERNEL_ERROR_ERRNO_FUNCTION_NOT_SUPPORTED;
 		}
 
-		if (!Memory::IsValidAddress(outdataPtr) || outlen < 0x800) {
-			WARN_LOG_REPORT(FILESYS, "sceIoIoctl: Invalid out pointer while reading ISO9660 volume descriptor");
+		if (!Memory::IsValidRange(outdataPtr, 0x800) || outlen < 0x800) {
+			WARN_LOG_REPORT(FILESYS, "sceIoIoctl: Invalid out pointer %08x while reading ISO9660 volume descriptor", outdataPtr);
 			return SCE_KERNEL_ERROR_ERRNO_INVALID_ARGUMENT;
 		}
 
 		INFO_LOG(SCEIO, "sceIoIoctl: reading ISO9660 volume descriptor read");
-		blockDevice->ReadBlock(16, Memory::GetPointer(outdataPtr));
+		blockDevice->ReadBlock(16, Memory::GetPointerWriteUnchecked(outdataPtr));
 		return 0;
 
 	// Get ISO9660 path table (from open ISO9660 file.)
@@ -437,12 +427,12 @@ int ISOFileSystem::Ioctl(u32 handle, u32 cmd, u32 indataPtr, u32 inlen, u32 outd
 
 		VolDescriptor desc;
 		blockDevice->ReadBlock(16, (u8 *)&desc);
-		if (outlen < (u32)desc.pathTableLengthLE) {
+		if (outlen < (u32)desc.pathTableLength) {
 			return SCE_KERNEL_ERROR_ERRNO_INVALID_ARGUMENT;
 		} else {
-			int block = (u16)desc.firstLETableSectorLE;
-			u32 size = (u32)desc.pathTableLengthLE;
-			u8 *out = Memory::GetPointer(outdataPtr);
+			int block = (u16)desc.firstLETableSector;
+			u32 size = Memory::ValidSize(outdataPtr, (u32)desc.pathTableLength);
+			u8 *out = Memory::GetPointerWriteRange(outdataPtr, size);
 
 			int blocks = size / blockDevice->GetBlockSize();
 			blockDevice->ReadBlocks(block, blocks, out);
@@ -461,10 +451,20 @@ int ISOFileSystem::Ioctl(u32 handle, u32 cmd, u32 indataPtr, u32 inlen, u32 outd
 	return SCE_KERNEL_ERROR_ERRNO_FUNCTION_NOT_SUPPORTED;
 }
 
-int ISOFileSystem::DevType(u32 handle)
-{
+PSPDevType ISOFileSystem::DevType(u32 handle) {
 	EntryMap::iterator iter = entries.find(handle);
-	return iter->second.isBlockSectorMode ? PSP_DEV_TYPE_BLOCK : PSP_DEV_TYPE_FILE;
+	if (iter == entries.end())
+		return PSPDevType::FILE;
+	PSPDevType type = iter->second.isBlockSectorMode ? PSPDevType::BLOCK : PSPDevType::FILE;
+	if (iter->second.isRawSector)
+		type |= PSPDevType::EMU_LBN;
+	return type;
+}
+
+FileSystemFlags ISOFileSystem::Flags() {
+	// TODO: Here may be a good place to force things, in case users recompress games
+	// as PBP or CSO when they were originally the other type.
+	return blockDevice->IsDisc() ? FileSystemFlags::UMD : FileSystemFlags::CARD;
 }
 
 size_t ISOFileSystem::ReadFile(u32 handle, u8 *pointer, s64 size)
@@ -515,7 +515,13 @@ size_t ISOFileSystem::ReadFile(u32 handle, u8 *pointer, s64 size, int &usec) {
 		if ((s64)e.seekPos + size > fileSize) {
 			// Clamp to the remaining size, but read what we can.
 			const s64 newSize = fileSize - (s64)e.seekPos;
-			WARN_LOG(FILESYS, "Reading beyond end of file, clamping size %lld to %lld", size, newSize);
+			// Reading beyond the file is really quite normal behavior (if return value handled correctly), so
+			// not doing WARN here. Still, can potentially be useful to see so leaving at INFO.
+			if (newSize == 0) {
+				INFO_LOG(FILESYS, "Attempted read at end of file, 0-size read simulated");
+			} else {
+				INFO_LOG(FILESYS, "Reading beyond end of file from seekPos %d, clamping size %lld to %lld", e.seekPos, size, newSize);
+			}
 			size = newSize;
 		}
 
@@ -527,7 +533,9 @@ size_t ISOFileSystem::ReadFile(u32 handle, u8 *pointer, s64 size, int &usec) {
 		u32 secNum = (u32)(positionOnIso / 2048);
 		u8 theSector[2048];
 
-		_dbg_assert_msg_(FILESYS, (middleSize & 2047) == 0, "Remaining size should be aligned");
+		if ((middleSize & 2047) != 0) {
+			ERROR_LOG(FILESYS, "Remaining size should be aligned");
+		}
 
 		const u8 *const start = pointer;
 		if (firstBlockSize > 0) {
@@ -607,7 +615,9 @@ PSPFileInfo ISOFileSystem::GetFileInfo(std::string filename) {
 		PSPFileInfo fileInfo;
 		fileInfo.name = filename;
 		fileInfo.exists = true;
+		fileInfo.type = FILETYPE_NORMAL;
 		fileInfo.size = readSize;
+		fileInfo.access = 0444;
 		fileInfo.startSector = sectorStart;
 		fileInfo.isOnSectorSystem = true;
 		fileInfo.numSectors = (readSize + sectorSize - 1) / sectorSize;
@@ -616,12 +626,10 @@ PSPFileInfo ISOFileSystem::GetFileInfo(std::string filename) {
 
 	TreeEntry *entry = GetFromPath(filename, false);
 	PSPFileInfo x; 
-	if (!entry) {
-		x.size = 0;
-		x.exists = false;
-	} else {
+	if (entry) {
 		x.name = entry->name;
-		x.access = FILEACCESS_READ;
+		// Strangely, it seems to be executable even for files.
+		x.access = 0555;
 		x.size = entry->size;
 		x.exists = true;
 		x.type = entry->isDirectory ? FILETYPE_DIRECTORY : FILETYPE_NORMAL;
@@ -631,11 +639,14 @@ PSPFileInfo ISOFileSystem::GetFileInfo(std::string filename) {
 	return x;
 }
 
-std::vector<PSPFileInfo> ISOFileSystem::GetDirListing(std::string path) {
+std::vector<PSPFileInfo> ISOFileSystem::GetDirListing(const std::string &path, bool *exists) {
 	std::vector<PSPFileInfo> myVector;
 	TreeEntry *entry = GetFromPath(path);
-	if (!entry)
+	if (!entry) {
+		if (exists)
+			*exists = false;
 		return myVector;
+	}
 
 	const std::string dot(".");
 	const std::string dotdot("..");
@@ -649,18 +660,19 @@ std::vector<PSPFileInfo> ISOFileSystem::GetDirListing(std::string path) {
 
 		PSPFileInfo x;
 		x.name = e->name;
-		x.access = FILEACCESS_READ;
+		// Strangely, it seems to be executable even for files.
+		x.access = 0555;
+		x.exists = true;
 		x.size = e->size;
 		x.type = e->isDirectory ? FILETYPE_DIRECTORY : FILETYPE_NORMAL;
 		x.isOnSectorSystem = true;
 		x.startSector = e->startingPosition/2048;
 		x.sectorSize = sectorSize;
 		x.numSectors = (u32)((e->size + sectorSize - 1) / sectorSize);
-		memset(&x.atime, 0, sizeof(x.atime));
-		memset(&x.mtime, 0, sizeof(x.mtime));
-		memset(&x.ctime, 0, sizeof(x.ctime));
 		myVector.push_back(x);
 	}
+	if (exists)
+		*exists = true;
 	return myVector;
 }
 
@@ -702,7 +714,7 @@ void ISOFileSystem::DoState(PointerWrap &p) {
 		return;
 
 	int n = (int) entries.size();
-	p.Do(n);
+	Do(p, n);
 
 	if (p.mode == p.MODE_READ) {
 		entries.clear();
@@ -710,18 +722,18 @@ void ISOFileSystem::DoState(PointerWrap &p) {
 			u32 fd = 0;
 			OpenFileEntry of;
 
-			p.Do(fd);
-			p.Do(of.seekPos);
-			p.Do(of.isRawSector);
-			p.Do(of.isBlockSectorMode);
-			p.Do(of.sectorStart);
-			p.Do(of.openSize);
+			Do(p, fd);
+			Do(p, of.seekPos);
+			Do(p, of.isRawSector);
+			Do(p, of.isBlockSectorMode);
+			Do(p, of.sectorStart);
+			Do(p, of.openSize);
 
 			bool hasFile = false;
-			p.Do(hasFile);
+			Do(p, hasFile);
 			if (hasFile) {
 				std::string path;
-				p.Do(path);
+				Do(p, path);
 				of.file = GetFromPath(path);
 			} else {
 				of.file = NULL;
@@ -732,24 +744,24 @@ void ISOFileSystem::DoState(PointerWrap &p) {
 	} else {
 		for (EntryMap::iterator it = entries.begin(), end = entries.end(); it != end; ++it) {
 			OpenFileEntry &of = it->second;
-			p.Do(it->first);
-			p.Do(of.seekPos);
-			p.Do(of.isRawSector);
-			p.Do(of.isBlockSectorMode);
-			p.Do(of.sectorStart);
-			p.Do(of.openSize);
+			Do(p, it->first);
+			Do(p, of.seekPos);
+			Do(p, of.isRawSector);
+			Do(p, of.isBlockSectorMode);
+			Do(p, of.sectorStart);
+			Do(p, of.openSize);
 
 			bool hasFile = of.file != NULL;
-			p.Do(hasFile);
+			Do(p, hasFile);
 			if (hasFile) {
 				std::string path = EntryFullPath(of.file);
-				p.Do(path);
+				Do(p, path);
 			}
 		}
 	}
 
 	if (s >= 2) {
-		p.Do(lastReadBlock_);
+		Do(p, lastReadBlock_);
 	} else {
 		lastReadBlock_ = 0;
 	}
