@@ -21,7 +21,8 @@
 // Some parts, especially in this file, were simply copied, so I guess this really makes this file GPL3.
 
 #include <algorithm>
-#include "Common/ChunkFile.h"
+#include "Common/Serialize/Serializer.h"
+#include "Common/Serialize/SerializeFuncs.h"
 #include "Core/MemMap.h"
 #include "Core/Reporting.h"
 #include "Core/Font/PGF.h"
@@ -37,10 +38,10 @@ static bool isJPCSPFont(const char *fontName) {
 
 // Gets a number of bits from an offset.
 static int getBits(int numBits, const u8 *buf, size_t pos) {
-	_dbg_assert_msg_(SCEFONT, numBits <= 32, "Unable to return more than 32 bits, %d requested", numBits);
+	_dbg_assert_msg_(numBits <= 32, "Unable to return more than 32 bits, %d requested", numBits);
 
 	const size_t wordpos = pos >> 5;
-	const u32 *wordbuf = (const u32 *)buf;
+	const u32_le *wordbuf = (const u32_le *)buf;
 	const u8 bitoff = pos & 31;
 
 	// Might just be in one, has to be within two.
@@ -52,8 +53,10 @@ static int getBits(int numBits, const u8 *buf, size_t pos) {
 
 		const u8 done = 32 - bitoff;
 		const u8 remaining = numBits - done;
-		const u32 mask = (1 << remaining) - 1;
-		v |= (wordbuf[wordpos + 1] & mask) << done;
+		if (remaining > 0) {
+			const u32 mask = (1 << remaining) - 1;
+			v |= (wordbuf[wordpos + 1] & mask) << done;
+		}
 		return v;
 	}
 }
@@ -128,12 +131,12 @@ void PGF::DoState(PointerWrap &p) {
 	if (!s)
 		return;
 
-	p.Do(header);
-	p.Do(rev3extra);
+	Do(p, header);
+	Do(p, rev3extra);
 
 	// Don't savestate size_t directly, 32-bit and 64-bit are different.
 	u32 fontDataSizeTemp = (u32)fontDataSize;
-	p.Do(fontDataSizeTemp);
+	Do(p, fontDataSizeTemp);
 	fontDataSize = (size_t)fontDataSizeTemp;
 	if (p.mode == p.MODE_READ) {
 		if (fontData) {
@@ -141,39 +144,39 @@ void PGF::DoState(PointerWrap &p) {
 		}
 		if (fontDataSize) {
 			fontData = new u8[fontDataSize];
-			p.DoArray(fontData, (int)fontDataSize);
+			DoArray(p, fontData, (int)fontDataSize);
 		}
 	} else if (fontDataSize) {
-		p.DoArray(fontData, (int)fontDataSize);
+		DoArray(p, fontData, (int)fontDataSize);
 	}
-	p.Do(fileName);
+	Do(p, fileName);
 
-	p.DoArray(dimensionTable, ARRAY_SIZE(dimensionTable));
-	p.DoArray(xAdjustTable, ARRAY_SIZE(xAdjustTable));
-	p.DoArray(yAdjustTable, ARRAY_SIZE(yAdjustTable));
-	p.DoArray(advanceTable, ARRAY_SIZE(advanceTable));
-	p.DoArray(charmapCompressionTable1, ARRAY_SIZE(charmapCompressionTable1));
-	p.DoArray(charmapCompressionTable2, ARRAY_SIZE(charmapCompressionTable2));
+	DoArray(p, dimensionTable, ARRAY_SIZE(dimensionTable));
+	DoArray(p, xAdjustTable, ARRAY_SIZE(xAdjustTable));
+	DoArray(p, yAdjustTable, ARRAY_SIZE(yAdjustTable));
+	DoArray(p, advanceTable, ARRAY_SIZE(advanceTable));
+	DoArray(p, charmapCompressionTable1, ARRAY_SIZE(charmapCompressionTable1));
+	DoArray(p, charmapCompressionTable2, ARRAY_SIZE(charmapCompressionTable2));
 
-	p.Do(charmap_compr);
-	p.Do(charmap);
+	Do(p, charmap_compr);
+	Do(p, charmap);
 	if (s == 1) {
 		std::vector<GlyphFromPGF1State> oldGlyphs;
-		p.Do(oldGlyphs);
+		Do(p, oldGlyphs);
 		glyphs.resize(oldGlyphs.size());
 		for (size_t i = 0; i < oldGlyphs.size(); ++i) {
 			glyphs[i] = oldGlyphs[i];
 		}
-		p.Do(oldGlyphs);
+		Do(p, oldGlyphs);
 		shadowGlyphs.resize(oldGlyphs.size());
 		for (size_t i = 0; i < oldGlyphs.size(); ++i) {
 			shadowGlyphs[i] = oldGlyphs[i];
 		}
 	} else {
-		p.Do(glyphs);
-		p.Do(shadowGlyphs);
+		Do(p, glyphs);
+		Do(p, shadowGlyphs);
 	}
-	p.Do(firstGlyph);
+	Do(p, firstGlyph);
 }
 
 bool PGF::ReadPtr(const u8 *ptr, size_t dataSize) {
@@ -227,14 +230,14 @@ bool PGF::ReadPtr(const u8 *ptr, size_t dataSize) {
 
 	const u8 *uptr = (const u8 *)wptr;
 
-	if (uptr >= startPtr + dataSize) {
-		return false;
-	}
-
 	int shadowCharMapSize = ((header.shadowMapLength * header.shadowMapBpe + 31) & ~31) / 8;
 	const u8 *shadowCharMap = uptr;
 	uptr += shadowCharMapSize;
 
+	if (uptr < startPtr || uptr >= startPtr + dataSize) {
+		return false;
+	}
+
 	const u16_le *sptr = (const u16_le *)uptr;
 	if (header.revision == 3) {
 		charmapCompressionTable1[0].resize(rev3extra.compCharMapLength1);
@@ -254,10 +257,6 @@ bool PGF::ReadPtr(const u8 *ptr, size_t dataSize) {
 
 	uptr = (const u8 *)sptr;
 
-	if (uptr >= startPtr + dataSize) {
-		return false;
-	}
-
 	int charMapSize = ((header.charMapLength * header.charMapBpe + 31) & ~31) / 8;
 	const u8 *charMap = uptr;
 	uptr += charMapSize;
@@ -266,6 +265,10 @@ bool PGF::ReadPtr(const u8 *ptr, size_t dataSize) {
 	const u8 *charPointerTable = uptr;
 	uptr += charPointerSize;
 
+	if (uptr < startPtr || uptr >= startPtr + dataSize) {
+		return false;
+	}
+
 	// PGF Fontdata.
 	u32 fontDataOffset = (u32)(uptr - startPtr);
 
@@ -347,7 +350,7 @@ bool PGF::GetCharInfo(int charCode, PGFCharInfo *charInfo, int altCharCode, int
 	charInfo->sfp26Height = glyph.dimensionHeight;
 	charInfo->sfp26Ascender = glyph.yAdjustH;
 	// Font y goes upwards.  If top is 10 and height is 11, the descender is approx. -1 (below 0.)
-	charInfo->sfp26Descender = charInfo->sfp26Ascender - (s32_le)charInfo->sfp26Height;
+	charInfo->sfp26Descender = charInfo->sfp26Ascender - (s32)charInfo->sfp26Height;
 	charInfo->sfp26BearingHX = glyph.xAdjustH;
 	charInfo->sfp26BearingHY = glyph.yAdjustH;
 	charInfo->sfp26BearingVX = glyph.xAdjustV;
@@ -676,6 +679,10 @@ void PGF::SetFontPixel(u32 base, int bpl, int bufWidth, int bufHeight, int x, in
 	}
 
 	static const u8 fontPixelSizeInBytes[] = { 0, 0, 1, 3, 4 }; // 0 means 2 pixels per byte
+	if (pixelformat < 0 || pixelformat > PSP_FONT_PIXELFORMAT_32) {
+		ERROR_LOG_REPORT_ONCE(pfgbadformat, SCEFONT, "Invalid image format in image: %d", (int)pixelformat);
+		return;
+	}
 	int pixelBytes = fontPixelSizeInBytes[pixelformat];
 	int bufMaxWidth = (pixelBytes == 0 ? bpl * 2 : bpl / pixelBytes);
 	if (x >= bufMaxWidth) {
