@@ -15,14 +15,16 @@
 // Official git repository and contact information can be found at
 // https://github.com/hrydgard/ppsspp and http://www.ppsspp.org/.
 
-#include "Common/ChunkFile.h"
+#include "Common/Serialize/Serializer.h"
+#include "Common/Serialize/SerializeFuncs.h"
+#include "Common/Data/Collections/FixedSizeQueue.h"
 #include "Core/MIPS/MIPS.h"
-#include "Core/Host.h"
 #include "Core/CoreTiming.h"
 #include "Core/HLE/HLE.h"
 #include "Core/HLE/FunctionWrappers.h"
 #include "Core/HLE/sceKernelThread.h"
 #include "Core/HLE/sceAudio.h"
+#include "Core/HLE/sceUsbMic.h"
 #include "Core/HLE/__sceAudio.h"
 #include "Core/Reporting.h"
 
@@ -33,24 +35,29 @@ const int AUDIO_ROUTING_SPEAKER_ON = 1;
 int defaultRoutingMode = AUDIO_ROUTING_SPEAKER_ON;
 int defaultRoutingVolMode = AUDIO_ROUTING_SPEAKER_ON;
 
+extern FixedSizeQueue<s16, 32768 * 8> chanSampleQueues[PSP_AUDIO_CHANNEL_MAX + 1];
+
+// The extra channel is for SRC/Output2/Vaudio.
+AudioChannel chans[PSP_AUDIO_CHANNEL_MAX + 1];
+
 void AudioChannel::DoState(PointerWrap &p)
 {
 	auto s = p.Section("AudioChannel", 1, 2);
 	if (!s)
 		return;
 
-	p.Do(reserved);
-	p.Do(sampleAddress);
-	p.Do(sampleCount);
-	p.Do(leftVolume);
-	p.Do(rightVolume);
-	p.Do(format);
-	p.Do(waitingThreads);
+	Do(p, reserved);
+	Do(p, sampleAddress);
+	Do(p, sampleCount);
+	Do(p, leftVolume);
+	Do(p, rightVolume);
+	Do(p, format);
+	Do(p, waitingThreads);
 	if (s >= 2) {
-		p.Do(defaultRoutingMode);
-		p.Do(defaultRoutingVolMode);
+		Do(p, defaultRoutingMode);
+		Do(p, defaultRoutingVolMode);
 	}
-	sampleQueue.DoState(p);
+	chanSampleQueues[index].DoState(p);
 }
 
 void AudioChannel::reset()
@@ -67,22 +74,15 @@ void AudioChannel::clear()
 	format = 0;
 	sampleAddress = 0;
 	sampleCount = 0;
-	sampleQueue.clear();
+	chanSampleQueues[index].clear();
 	waitingThreads.clear();
 }
 
-// There's a second Audio api called Audio2 that only has one channel, I guess the 8 channel api was overkill.
-// We simply map it to an extra channel after the 8 channels, since they can be used concurrently.
-
-// The extra channel is for SRC/Output2/Vaudio.
-AudioChannel chans[PSP_AUDIO_CHANNEL_MAX + 1];
-
-// Enqueues the buffer pointer on the channel. If channel buffer queue is full (2 items?) will block until it isn't.
-// For solid audio output we'll need a queue length of 2 buffers at least, we'll try that first.
+// Enqueues the buffer pointed to on the channel. If channel buffer queue is full (2 items?) will block until it isn't.
+// For solid audio output we'll need a queue length of 2 buffers at least.
 
 // Not sure about the range of volume, I often see 0x800 so that might be either
 // max or 50%?
-
 static u32 sceAudioOutputBlocking(u32 chan, int vol, u32 samplePtr) {
 	if (vol > 0xFFFF) {
 		ERROR_LOG(SCEAUDIO, "sceAudioOutputBlocking() - invalid volume");
@@ -181,7 +181,7 @@ static int sceAudioGetChannelRestLen(u32 chan) {
 		ERROR_LOG(SCEAUDIO, "sceAudioGetChannelRestLen(%08x) - bad channel", chan);
 		return SCE_ERROR_AUDIO_INVALID_CHANNEL;
 	}
-	int remainingSamples = (int)chans[chan].sampleQueue.size() / 2;
+	int remainingSamples = (int)chanSampleQueues[chan].size() / 2;
 	VERBOSE_LOG(SCEAUDIO, "%d=sceAudioGetChannelRestLen(%08x)", remainingSamples, chan);
 	return remainingSamples;
 }
@@ -191,7 +191,7 @@ static int sceAudioGetChannelRestLength(u32 chan) {
 		ERROR_LOG(SCEAUDIO, "sceAudioGetChannelRestLength(%08x) - bad channel", chan);
 		return SCE_ERROR_AUDIO_INVALID_CHANNEL;
 	}
-	int remainingSamples = (int)chans[chan].sampleQueue.size() / 2;
+	int remainingSamples = (int)chanSampleQueues[chan].size() / 2;
 	VERBOSE_LOG(SCEAUDIO, "%d=sceAudioGetChannelRestLength(%08x)", remainingSamples, chan);
 	return remainingSamples;
 }
@@ -248,10 +248,10 @@ static u32 sceAudioChRelease(u32 chan) {
 		ERROR_LOG(SCEAUDIO, "sceAudioChRelease(%i) - channel not reserved", chan);
 		return SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED;
 	}
-	DEBUG_LOG(SCEAUDIO, "sceAudioChRelease(%i)", chan);
+	// TODO: Does this error if busy?
 	chans[chan].reset();
 	chans[chan].reserved = false;
-	return 1;
+	return hleLogSuccessI(SCEAUDIO, 0);
 }
 
 static u32 sceAudioSetChannelDataLen(u32 chan, u32 len) {
@@ -317,61 +317,80 @@ static u32 sceAudioEnd() {
 }
 
 static u32 sceAudioOutput2Reserve(u32 sampleCount) {
+	auto &chan = chans[PSP_AUDIO_CHANNEL_OUTPUT2];
+	// This seems to ignore the MSB, for some reason.
+	sampleCount &= 0x7FFFFFFF;
 	if (sampleCount < 17 || sampleCount > 4111) {
-		ERROR_LOG(SCEAUDIO, "sceAudioOutput2Reserve(%08x) - invalid sample count", sampleCount);
-		return SCE_KERNEL_ERROR_INVALID_SIZE;
-	} else if (chans[PSP_AUDIO_CHANNEL_OUTPUT2].reserved) {
-		ERROR_LOG(SCEAUDIO, "sceAudioOutput2Reserve(%08x) - channel already reserved", sampleCount);
-		return SCE_ERROR_AUDIO_CHANNEL_ALREADY_RESERVED;
-	} else {
-		DEBUG_LOG(SCEAUDIO, "sceAudioOutput2Reserve(%08x)", sampleCount);
-		chans[PSP_AUDIO_CHANNEL_OUTPUT2].sampleCount = sampleCount;
-		chans[PSP_AUDIO_CHANNEL_OUTPUT2].format = PSP_AUDIO_FORMAT_STEREO;
-		chans[PSP_AUDIO_CHANNEL_OUTPUT2].reserved = true;
+		return hleLogError(SCEAUDIO, SCE_KERNEL_ERROR_INVALID_SIZE, "invalid sample count");
+	} else if (chan.reserved) {
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_ALREADY_RESERVED, "channel already reserved");
 	}
-	return 0;
+
+	chan.sampleCount = sampleCount;
+	chan.format = PSP_AUDIO_FORMAT_STEREO;
+	chan.reserved = true;
+	__AudioSetSRCFrequency(0);
+	return hleLogSuccessI(SCEAUDIO, 0);
 }
 
 static u32 sceAudioOutput2OutputBlocking(u32 vol, u32 dataPtr) {
 	// Note: 0xFFFFF, not 0xFFFF!
 	if (vol > 0xFFFFF) {
-		ERROR_LOG(SCEAUDIO, "sceAudioOutput2OutputBlocking(%08x, %08x) - invalid volume", vol, dataPtr);
-		return SCE_ERROR_AUDIO_INVALID_VOLUME;
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_INVALID_VOLUME, "invalid volume");
 	}
-	DEBUG_LOG(SCEAUDIO, "sceAudioOutput2OutputBlocking(%08x, %08x)", vol, dataPtr);
-	chans[PSP_AUDIO_CHANNEL_OUTPUT2].leftVolume = vol;
-	chans[PSP_AUDIO_CHANNEL_OUTPUT2].rightVolume = vol;
-	chans[PSP_AUDIO_CHANNEL_OUTPUT2].sampleAddress = dataPtr;
-	return __AudioEnqueue(chans[PSP_AUDIO_CHANNEL_OUTPUT2], PSP_AUDIO_CHANNEL_OUTPUT2, true);
+
+	auto &chan = chans[PSP_AUDIO_CHANNEL_OUTPUT2];
+	if (!chan.reserved) {
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED, "channel not reserved");
+	}
+
+	chan.leftVolume = vol;
+	chan.rightVolume = vol;
+	chan.sampleAddress = dataPtr;
+
+	hleEatCycles(10000);
+	int result = __AudioEnqueue(chan, PSP_AUDIO_CHANNEL_OUTPUT2, true);
+	if (result < 0)
+		return hleLogError(SCEAUDIO, result);
+	return hleLogSuccessI(SCEAUDIO, result);
 }
 
 static u32 sceAudioOutput2ChangeLength(u32 sampleCount) {
-	if (!chans[PSP_AUDIO_CHANNEL_OUTPUT2].reserved) {
-		ERROR_LOG(SCEAUDIO, "sceAudioOutput2ChangeLength(%08x) - channel not reserved ", sampleCount);
-		return SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED;
+	auto &chan = chans[PSP_AUDIO_CHANNEL_OUTPUT2];
+	if (!chan.reserved) {
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED, "channel not reserved");
 	}
-	DEBUG_LOG(SCEAUDIO, "sceAudioOutput2ChangeLength(%08x)", sampleCount);
-	chans[PSP_AUDIO_CHANNEL_OUTPUT2].sampleCount = sampleCount;
-	return 0;
+	chan.sampleCount = sampleCount;
+	return hleLogSuccessI(SCEAUDIO, 0);
 }
 
 static u32 sceAudioOutput2GetRestSample() {
-	if (!chans[PSP_AUDIO_CHANNEL_OUTPUT2].reserved) {
-		ERROR_LOG(SCEAUDIO, "sceAudioOutput2GetRestSample() - channel not reserved ");
-		return SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED;
+	auto &chan = chans[PSP_AUDIO_CHANNEL_OUTPUT2];
+	if (!chan.reserved) {
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED, "channel not reserved");
 	}
-	DEBUG_LOG(SCEAUDIO, "sceAudioOutput2GetRestSample()");
-	return (u32) chans[PSP_AUDIO_CHANNEL_OUTPUT2].sampleQueue.size() / 2;
+	u32 size = (u32)chanSampleQueues[PSP_AUDIO_CHANNEL_OUTPUT2].size() / 2;
+	if (size > chan.sampleCount) {
+		// If ChangeLength reduces the size, it still gets output but this return is clamped.
+		size = chan.sampleCount;
+	}
+	return hleLogSuccessI(SCEAUDIO, size);
 }
 
 static u32 sceAudioOutput2Release() {
-	DEBUG_LOG(SCEAUDIO, "sceAudioOutput2Release()");
-	chans[PSP_AUDIO_CHANNEL_OUTPUT2].reset();
-	chans[PSP_AUDIO_CHANNEL_OUTPUT2].reserved = false;
-	return 0;
+	auto &chan = chans[PSP_AUDIO_CHANNEL_OUTPUT2];
+	if (!chan.reserved)
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED, "channel not reserved");
+	if (!chanSampleQueues[PSP_AUDIO_CHANNEL_OUTPUT2].empty())
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_ALREADY_RESERVED, "output busy");
+
+	chan.reset();
+	chan.reserved = false;
+	return hleLogSuccessI(SCEAUDIO, 0);
 }
 
 static u32 sceAudioSetFrequency(u32 freq) {
+	// TODO: Not available from user code.
 	if (freq == 44100 || freq == 48000) {
 		INFO_LOG(SCEAUDIO, "sceAudioSetFrequency(%08x)", freq);
 		__AudioSetOutputFrequency(freq);
@@ -387,50 +406,95 @@ static u32 sceAudioSetVolumeOffset() {
 	return 0;
 }
 
+static bool SRCFrequencyAllowed(int freq) {
+	if (freq == 44100 || freq == 22050 || freq == 11025)
+		return true;
+	if (freq == 48000 || freq == 32000 || freq == 24000 || freq == 16000 || freq == 12000 || freq == 8000)
+		return true;
+	return false;
+}
+
 static u32 sceAudioSRCChReserve(u32 sampleCount, u32 freq, u32 format) {
+	auto &chan = chans[PSP_AUDIO_CHANNEL_SRC];
+	// This seems to ignore the MSB, for some reason.
+	sampleCount &= 0x7FFFFFFF;
 	if (format == 4) {
-		ERROR_LOG(SCEAUDIO, "sceAudioSRCChReserve(%08x, %08x, %08x) - unexpected format", sampleCount, freq, format);
-		return PSP_AUDIO_ERROR_SRC_FORMAT_4;
+		return hleReportError(SCEAUDIO, PSP_AUDIO_ERROR_SRC_FORMAT_4, "unexpected format");
 	} else if (format != 2) {
-		ERROR_LOG(SCEAUDIO, "sceAudioSRCChReserve(%08x, %08x, %08x) - unexpected format", sampleCount, freq, format);
-		return SCE_KERNEL_ERROR_INVALID_SIZE;
+		return hleLogError(SCEAUDIO, SCE_KERNEL_ERROR_INVALID_SIZE, "unexpected format");
 	} else if (sampleCount < 17 || sampleCount > 4111) {
-		ERROR_LOG(SCEAUDIO, "sceAudioSRCChReserve(%08x, %08x, %08x) - invalid sample count", sampleCount, freq, format);
-		return SCE_KERNEL_ERROR_INVALID_SIZE;
-	} else if (chans[PSP_AUDIO_CHANNEL_SRC].reserved) {
-		ERROR_LOG(SCEAUDIO, "sceAudioSRCChReserve(%08x, %08x, %08x) - channel already reserved ", sampleCount, freq, format);
-		return SCE_ERROR_AUDIO_CHANNEL_ALREADY_RESERVED;
-	} else {
-		DEBUG_LOG(SCEAUDIO, "sceAudioSRCChReserve(%08x, %08x, %08x)", sampleCount, freq, format);
-		chans[PSP_AUDIO_CHANNEL_SRC].reserved = true;
-		chans[PSP_AUDIO_CHANNEL_SRC].sampleCount = sampleCount;
-		chans[PSP_AUDIO_CHANNEL_SRC].format = format == 2 ? PSP_AUDIO_FORMAT_STEREO : PSP_AUDIO_FORMAT_MONO;
-		__AudioSetOutputFrequency(freq);
+		return hleLogError(SCEAUDIO, SCE_KERNEL_ERROR_INVALID_SIZE, "invalid sample count");
+	} else if (freq != 0 && !SRCFrequencyAllowed(freq)) {
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_INVALID_FREQUENCY, "invalid frequency");
+	} else if (chan.reserved) {
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_ALREADY_RESERVED, "channel already reserved");
 	}
-	return 0;
+
+	chan.reserved = true;
+	chan.sampleCount = sampleCount;
+	chan.format = format == 2 ? PSP_AUDIO_FORMAT_STEREO : PSP_AUDIO_FORMAT_MONO;
+	// Zero means default to 44.1kHz.
+	__AudioSetSRCFrequency(freq);
+	return hleLogSuccessI(SCEAUDIO, 0);
 }
 
 static u32 sceAudioSRCChRelease() {
-	if (!chans[PSP_AUDIO_CHANNEL_SRC].reserved) {
-		ERROR_LOG(SCEAUDIO, "sceAudioSRCChRelease() - channel not reserved ");
-		return SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED;
-	}
-	DEBUG_LOG(SCEAUDIO, "sceAudioSRCChRelease()");
-	chans[PSP_AUDIO_CHANNEL_SRC].reset();
-	chans[PSP_AUDIO_CHANNEL_SRC].reserved = false;
-	return 0;
+	auto &chan = chans[PSP_AUDIO_CHANNEL_SRC];
+	if (!chan.reserved)
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED, "channel not reserved");
+	if (!chanSampleQueues[PSP_AUDIO_CHANNEL_SRC].empty())
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_ALREADY_RESERVED, "output busy");
+
+	chan.reset();
+	chan.reserved = false;
+	return hleLogSuccessI(SCEAUDIO, 0);
 }
 
 static u32 sceAudioSRCOutputBlocking(u32 vol, u32 buf) {
 	if (vol > 0xFFFFF) {
-		ERROR_LOG(SCEAUDIO, "sceAudioSRCOutputBlocking(%08x, %08x) - invalid volume", vol, buf);
-		return SCE_ERROR_AUDIO_INVALID_VOLUME;
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_INVALID_VOLUME, "invalid volume");
 	}
-	DEBUG_LOG(SCEAUDIO, "sceAudioSRCOutputBlocking(%08x, %08x)", vol, buf);
-	chans[PSP_AUDIO_CHANNEL_SRC].leftVolume = vol;
-	chans[PSP_AUDIO_CHANNEL_SRC].rightVolume = vol;
-	chans[PSP_AUDIO_CHANNEL_SRC].sampleAddress = buf;
-	return __AudioEnqueue(chans[PSP_AUDIO_CHANNEL_SRC], PSP_AUDIO_CHANNEL_SRC, true);
+
+	auto &chan = chans[PSP_AUDIO_CHANNEL_SRC];
+	if (!chan.reserved) {
+		return hleLogError(SCEAUDIO, SCE_ERROR_AUDIO_CHANNEL_NOT_RESERVED, "channel not reserved");
+	}
+
+	chan.leftVolume = vol;
+	chan.rightVolume = vol;
+	chan.sampleAddress = buf;
+
+	hleEatCycles(10000);
+	int result = __AudioEnqueue(chan, PSP_AUDIO_CHANNEL_SRC, true);
+	if (result < 0)
+		return hleLogError(SCEAUDIO, result);
+	return hleLogSuccessI(SCEAUDIO, result);
+}
+
+static int sceAudioInputBlocking(u32 maxSamples, u32 sampleRate, u32 bufAddr) {
+	if (!Memory::IsValidAddress(bufAddr)) {
+		ERROR_LOG(HLE, "sceAudioInputBlocking(%d, %d, %08x): invalid addresses", maxSamples, sampleRate, bufAddr);
+		return -1;
+	}
+
+	INFO_LOG(HLE, "sceAudioInputBlocking: maxSamples: %d, samplerate: %d, bufAddr: %08x", maxSamples, sampleRate, bufAddr);
+	return __MicInput(maxSamples, sampleRate, bufAddr, AUDIOINPUT);
+}
+
+static int sceAudioInput(u32 maxSamples, u32 sampleRate, u32 bufAddr) {
+	if (!Memory::IsValidAddress(bufAddr)) {
+		ERROR_LOG(HLE, "sceAudioInput(%d, %d, %08x): invalid addresses", maxSamples, sampleRate, bufAddr);
+		return -1;
+	}
+
+	ERROR_LOG(HLE, "UNTEST sceAudioInput: maxSamples: %d, samplerate: %d, bufAddr: %08x", maxSamples, sampleRate, bufAddr);
+	return __MicInput(maxSamples, sampleRate, bufAddr, AUDIOINPUT, false);
+}
+
+static int sceAudioGetInputLength() {
+	int ret = Microphone::getReadMicDataLength() / 2;
+	ERROR_LOG(HLE, "UNTEST sceAudioGetInputLength(ret: %d)", ret);
+	return ret;
 }
 
 static u32 sceAudioRoutingSetMode(u32 mode) {
@@ -461,31 +525,31 @@ const HLEFunction sceAudio[] =
 {
 	// Newer simplified single channel audio output. Presumably for games that use Atrac3
 	// directly from Sas instead of playing it on a separate audio channel.
-	{0X01562BA3, &WrapU_U<sceAudioOutput2Reserve>,          "sceAudioOutput2Reserve",        'x', "x"   },
+	{0X01562BA3, &WrapU_U<sceAudioOutput2Reserve>,          "sceAudioOutput2Reserve",        'x', "i"   },
 	{0X2D53F36E, &WrapU_UU<sceAudioOutput2OutputBlocking>,  "sceAudioOutput2OutputBlocking", 'x', "xx"  },
-	{0X63F2889C, &WrapU_U<sceAudioOutput2ChangeLength>,     "sceAudioOutput2ChangeLength",   'x', "x"   },
-	{0X647CEF33, &WrapU_V<sceAudioOutput2GetRestSample>,    "sceAudioOutput2GetRestSample",  'x', ""    },
+	{0X63F2889C, &WrapU_U<sceAudioOutput2ChangeLength>,     "sceAudioOutput2ChangeLength",   'x', "i"   },
+	{0X647CEF33, &WrapU_V<sceAudioOutput2GetRestSample>,    "sceAudioOutput2GetRestSample",  'i', ""    },
 	{0X43196845, &WrapU_V<sceAudioOutput2Release>,          "sceAudioOutput2Release",        'x', ""    },
 
 	// "Traditional" audio channel interface
 	{0X80F1F7E0, &WrapU_V<sceAudioInit>,                    "sceAudioInit",                  'x', ""    },
 	{0X210567F7, &WrapU_V<sceAudioEnd>,                     "sceAudioEnd",                   'x', ""    },
-	{0XA2BEAA6C, &WrapU_U<sceAudioSetFrequency>,            "sceAudioSetFrequency",          'x', "x"   },
+	{0XA2BEAA6C, &WrapU_U<sceAudioSetFrequency>,            "sceAudioSetFrequency",          'x', "i"   },
 	{0X927AC32B, &WrapU_V<sceAudioSetVolumeOffset>,         "sceAudioSetVolumeOffset",       'x', ""    },
-	{0X8C1009B2, &WrapU_UIU<sceAudioOutput>,                "sceAudioOutput",                'x', "xix" },
-	{0X136CAF51, &WrapU_UIU<sceAudioOutputBlocking>,        "sceAudioOutputBlocking",        'x', "xix" },
-	{0XE2D56B2D, &WrapU_UIIU<sceAudioOutputPanned>,         "sceAudioOutputPanned",          'x', "xiix"},
-	{0X13F592BC, &WrapU_UIIU<sceAudioOutputPannedBlocking>, "sceAudioOutputPannedBlocking",  'x', "xiix"},
-	{0X5EC81C55, &WrapU_IUU<sceAudioChReserve>,             "sceAudioChReserve",             'x', "ixx" },
-	{0X6FC46853, &WrapU_U<sceAudioChRelease>,               "sceAudioChRelease",             'x', "x"   },
-	{0XE9D97901, &WrapI_U<sceAudioGetChannelRestLen>,       "sceAudioGetChannelRestLen",     'i', "x"   },
-	{0XB011922F, &WrapI_U<sceAudioGetChannelRestLength>,    "sceAudioGetChannelRestLength",  'i', "x"   },
-	{0XCB2E439E, &WrapU_UU<sceAudioSetChannelDataLen>,      "sceAudioSetChannelDataLen",     'x', "xx"  },
-	{0X95FD0C2D, &WrapU_UU<sceAudioChangeChannelConfig>,    "sceAudioChangeChannelConfig",   'x', "xx"  },
-	{0XB7E1D8E7, &WrapU_UUU<sceAudioChangeChannelVolume>,   "sceAudioChangeChannelVolume",   'x', "xxx" },
+	{0X8C1009B2, &WrapU_UIU<sceAudioOutput>,                "sceAudioOutput",                'x', "ixx" },
+	{0X136CAF51, &WrapU_UIU<sceAudioOutputBlocking>,        "sceAudioOutputBlocking",        'x', "ixx" },
+	{0XE2D56B2D, &WrapU_UIIU<sceAudioOutputPanned>,         "sceAudioOutputPanned",          'x', "ixxx"},
+	{0X13F592BC, &WrapU_UIIU<sceAudioOutputPannedBlocking>, "sceAudioOutputPannedBlocking",  'x', "ixxx"},
+	{0X5EC81C55, &WrapU_IUU<sceAudioChReserve>,             "sceAudioChReserve",             'x', "iii" },
+	{0X6FC46853, &WrapU_U<sceAudioChRelease>,               "sceAudioChRelease",             'x', "i"   },
+	{0XE9D97901, &WrapI_U<sceAudioGetChannelRestLen>,       "sceAudioGetChannelRestLen",     'i', "i"   },
+	{0XB011922F, &WrapI_U<sceAudioGetChannelRestLength>,    "sceAudioGetChannelRestLength",  'i', "i"   },
+	{0XCB2E439E, &WrapU_UU<sceAudioSetChannelDataLen>,      "sceAudioSetChannelDataLen",     'x', "ii"  },
+	{0X95FD0C2D, &WrapU_UU<sceAudioChangeChannelConfig>,    "sceAudioChangeChannelConfig",   'x', "ii"  },
+	{0XB7E1D8E7, &WrapU_UUU<sceAudioChangeChannelVolume>,   "sceAudioChangeChannelVolume",   'x', "ixx" },
 
-	// Not sure about the point of these, maybe like traditional but with ability to do sample rate conversion?
-	{0X38553111, &WrapU_UUU<sceAudioSRCChReserve>,          "sceAudioSRCChReserve",          'x', "xxx" },
+	// Like Output2, but with ability to do sample rate conversion.
+	{0X38553111, &WrapU_UUU<sceAudioSRCChReserve>,          "sceAudioSRCChReserve",          'x', "iii" },
 	{0X5C37C0AE, &WrapU_V<sceAudioSRCChRelease>,            "sceAudioSRCChRelease",          'x', ""    },
 	{0XE0727056, &WrapU_UU<sceAudioSRCOutputBlocking>,      "sceAudioSRCOutputBlocking",     'x', "xx"  },
 
@@ -496,9 +560,9 @@ const HLEFunction sceAudio[] =
 	// Microphone interface
 	{0X7DE61688, nullptr,                                   "sceAudioInputInit",             '?', ""    },
 	{0XE926D3FB, nullptr,                                   "sceAudioInputInitEx",           '?', ""    },
-	{0X6D4BEC68, nullptr,                                   "sceAudioInput",                 '?', ""    },
-	{0X086E5895, nullptr,                                   "sceAudioInputBlocking",         '?', ""    },
-	{0XA708C6A6, nullptr,                                   "sceAudioGetInputLength",        '?', ""    },
+	{0X6D4BEC68, &WrapI_UUU<sceAudioInput>,                 "sceAudioInput",                 'i', "xxx" },
+	{0X086E5895, &WrapI_UUU<sceAudioInputBlocking>,         "sceAudioInputBlocking",         'i', "xxx" },
+	{0XA708C6A6, &WrapI_V<sceAudioGetInputLength>,          "sceAudioGetInputLength",        'i', ""    },
 	{0XA633048E, nullptr,                                   "sceAudioPollInputEnd",          '?', ""    },
 	{0X87B2E651, nullptr,                                   "sceAudioWaitInputEnd",          '?', ""    },
 
