@@ -32,9 +32,9 @@
 #include <mutex>
 #include <condition_variable>
 
-#include "base/basictypes.h"
-#include "profiler/profiler.h"
-#include "thread/threadutil.h"
+#include "Common/Profiler/Profiler.h"
+#include "Common/Thread/ThreadUtil.h"
+#include "Common/Serialize/SerializeFuncs.h"
 #include "Common/Log.h"
 #include "Core/Config.h"
 #include "Core/CoreTiming.h"
@@ -101,7 +101,7 @@ static SasThreadParams sasThreadParams;
 static int sasMixEvent = -1;
 
 int __SasThread() {
-	setCurrentThreadName("SAS");
+	SetCurrentThreadName("SAS");
 
 	std::unique_lock<std::mutex> guard(sasWakeMutex);
 	while (sasThreadState != SasThreadState::DISABLED) {
@@ -109,10 +109,9 @@ int __SasThread() {
 		if (sasThreadState == SasThreadState::QUEUED) {
 			sas->Mix(sasThreadParams.outAddr, sasThreadParams.inAddr, sasThreadParams.leftVol, sasThreadParams.rightVol);
 
-			sasDoneMutex.lock();
+			std::lock_guard<std::mutex> doneGuard(sasDoneMutex);
 			sasThreadState = SasThreadState::READY;
 			sasDone.notify_one();
-			sasDoneMutex.unlock();
 		}
 	}
 	return 0;
@@ -204,18 +203,16 @@ void __SasDoState(PointerWrap &p) {
 		__SasDrain();
 	}
 
-	p.DoClass(sas);
+	DoClass(p, sas);
 
 	if (s >= 2) {
-		p.Do(sasMixEvent);
+		Do(p, sasMixEvent);
 	} else {
 		sasMixEvent = -1;
 		__SasDisableThread();
 	}
 
-	if (sasMixEvent != -1) {
-		CoreTiming::RestoreRegisterEvent(sasMixEvent, "SasMix", sasMixFinish);
-	}
+	CoreTiming::RestoreRegisterEvent(sasMixEvent, "SasMix", sasMixFinish);
 }
 
 void __SasShutdown() {
@@ -269,7 +266,7 @@ static u32 sceSasGetEndFlag(u32 core) {
 			endFlag |= (1 << i);
 	}
 
-	DEBUG_LOG(SCESAS, "%08x=sceSasGetEndFlag(%08x)", endFlag, core);
+	VERBOSE_LOG(SCESAS, "%08x=sceSasGetEndFlag(%08x)", endFlag, core);
 	return endFlag;
 }
 
@@ -299,7 +296,7 @@ static u32 _sceSasCore(u32 core, u32 outAddr) {
 
 	__SasEnqueueMix(outAddr);
 
-	return hleLogSuccessI(SCESAS, delaySasResult(0));
+	return hleLogSuccessVerboseI(SCESAS, delaySasResult(0));
 }
 
 // Another way of running the mixer, the inoutAddr should be both input and output
@@ -318,12 +315,12 @@ static u32 _sceSasCoreWithMix(u32 core, u32 inoutAddr, int leftVolume, int right
 
 	__SasEnqueueMix(inoutAddr, inoutAddr, leftVolume, rightVolume);
 
-	return hleLogSuccessI(SCESAS, delaySasResult(0));
+	return hleLogSuccessVerboseI(SCESAS, delaySasResult(0));
 }
 
 static u32 sceSasSetVoice(u32 core, int voiceNum, u32 vagAddr, int size, int loop) {
 	if (voiceNum >= PSP_SAS_VOICES_MAX || voiceNum < 0)	{
-		return hleLogWarning(SCESAS, ERROR_SAS_INVALID_VOICE, "invalid voicenum");
+		return hleLogVerbose(SCESAS, ERROR_SAS_INVALID_VOICE, "invalid voicenum");
 	}
 
 	if (size == 0 || ((u32)size & 0xF) != 0) {
@@ -366,8 +363,11 @@ static u32 sceSasSetVoice(u32 core, int voiceNum, u32 vagAddr, int size, int loo
 	v.type = VOICETYPE_VAG;
 	v.vagAddr = vagAddr;  // Real VAG header is 0x30 bytes behind the vagAddr
 	v.vagSize = size;
-	v.loop = loop ? true : false;
-	v.ChangedParams(vagAddr == prevVagAddr);
+	v.loop = loop != 0;
+	if (v.on) {
+		v.playing = true;
+	}
+	v.vag.Start(vagAddr, size, loop != 0);
 	return 0;
 }
 
@@ -404,7 +404,6 @@ static u32 sceSasSetVoicePCM(u32 core, int voiceNum, u32 pcmAddr, int size, int
 	v.pcmLoopPos = loopPos >= 0 ? loopPos : 0;
 	v.loop = loopPos >= 0 ? true : false;
 	v.playing = true;
-	v.ChangedParams(pcmAddr == prevPcmAddr);
 	return 0;
 }
 
@@ -427,7 +426,7 @@ static u32 sceSasSetPause(u32 core, u32 voicebit, int pause) {
 	for (int i = 0; voicebit != 0; i++, voicebit >>= 1) {
 		if (i < PSP_SAS_VOICES_MAX && i >= 0) {
 			if ((voicebit & 1) != 0)
-				sas->voices[i].paused = pause ? true : false;
+				sas->voices[i].paused = pause != 0;
 		}
 	}
 
@@ -469,7 +468,6 @@ static u32 sceSasSetPitch(u32 core, int voiceNum, int pitch) {
 	__SasDrain();
 	SasVoice &v = sas->voices[voiceNum];
 	v.pitch = pitch;
-	v.ChangedParams(false);
 	return 0;
 }
 
@@ -526,7 +524,6 @@ static u32 sceSasSetNoise(u32 core, int voiceNum, int freq) {
 	SasVoice &v = sas->voices[voiceNum];
 	v.type = VOICETYPE_NOISE;
 	v.noiseFreq = freq;
-	v.ChangedParams(true);
 	return 0;
 }
 
@@ -539,7 +536,7 @@ static u32 sceSasSetSL(u32 core, int voiceNum, int level) {
 	DEBUG_LOG(SCESAS, "sceSasSetSL(%08x, %i, %08x)", core, voiceNum, level);
 	__SasDrain();
 	SasVoice &v = sas->voices[voiceNum];
-	v.envelope.sustainLevel = level;
+	v.envelope.SetSustainLevel(level);
 	return 0;
 }
 
@@ -559,10 +556,7 @@ static u32 sceSasSetADSR(u32 core, int voiceNum, int flag, int a, int d, int s,
 
 	__SasDrain();
 	SasVoice &v = sas->voices[voiceNum];
-	if ((flag & 0x1) != 0) v.envelope.attackRate  = a;
-	if ((flag & 0x2) != 0) v.envelope.decayRate   = d;
-	if ((flag & 0x4) != 0) v.envelope.sustainRate = s;
-	if ((flag & 0x8) != 0) v.envelope.releaseRate = r;
+	v.envelope.SetRate(flag, a, d, s, r);
 	return 0;
 }
 
@@ -605,10 +599,7 @@ static u32 sceSasSetADSRMode(u32 core, int voiceNum, int flag, int a, int d, int
 	DEBUG_LOG(SCESAS, "sceSasSetADSRMode(%08x, %i, %i, %08x, %08x, %08x, %08x)", core, voiceNum, flag, a, d, s, r);
 	__SasDrain();
 	SasVoice &v = sas->voices[voiceNum];
-	if ((flag & 0x1) != 0) v.envelope.attackType  = a;
-	if ((flag & 0x2) != 0) v.envelope.decayType   = d;
-	if ((flag & 0x4) != 0) v.envelope.sustainType = s;
-	if ((flag & 0x8) != 0) v.envelope.releaseType = r;
+	v.envelope.SetEnvelope(flag, a, d, s, r);
 	return 0;
 }
 
