@@ -17,25 +17,30 @@
 
 #include <functional>
 
-#include "base/basictypes.h"
-#include "json/json_reader.h"
-
-#include "i18n/i18n.h"
-#include "ui/screen.h"
-#include "ui/ui_context.h"
-#include "ui/viewgroup.h"
-#include "gfx_es2/draw_buffer.h"
+#include "Common/UI/Screen.h"
+#include "Common/UI/Context.h"
+#include "Common/UI/ViewGroup.h"
+#include "Common/UI/IconCache.h"
+#include "Common/Render/DrawBuffer.h"
 
 #include "Common/Log.h"
+#include "Common/Data/Text/I18n.h"
+#include "Common/Data/Format/JSONReader.h"
 #include "Common/StringUtils.h"
+#include "Common/Render/ManagedTexture.h"
+#include "Common/Net/NetBuffer.h"
 #include "Core/Config.h"
 #include "Core/System.h"
 #include "Core/Util/GameManager.h"
 #include "UI/EmuScreen.h"
 #include "UI/Store.h"
-#include "UI/TextureUtil.h"
 
-const std::string storeBaseUrl = "http://store.ppsspp.org/";
+const char *storeBaseUrlHttp = "http://store.ppsspp.org/";
+const char *storeBaseUrlHttps = "https://store.ppsspp.org/";
+
+static std::string StoreBaseUrl() {
+	return System_GetPropertyBool(SYSPROP_SUPPORTS_HTTPS) ? storeBaseUrlHttps : storeBaseUrlHttp;
+}
 
 // baseUrl is assumed to have a trailing slash, and not contain any subdirectories.
 std::string ResolveUrl(std::string baseUrl, std::string url) {
@@ -43,7 +48,7 @@ std::string ResolveUrl(std::string baseUrl, std::string url) {
 		return baseUrl;
 	} else if (url[0] == '/') {
 		return baseUrl + url.substr(1);
-	} else if (url.substr(0, 7) == "http://") {
+	} else if (startsWith(url, "http://") || startsWith(url, "https://")) {
 		return url;
 	} else {
 		// Huh.
@@ -53,16 +58,38 @@ std::string ResolveUrl(std::string baseUrl, std::string url) {
 
 class HttpImageFileView : public UI::View {
 public:
-	HttpImageFileView(http::Downloader *downloader, const std::string &path, UI::ImageSizeMode sizeMode = UI::IS_DEFAULT, UI::LayoutParams *layoutParams = 0)
-		: UI::View(layoutParams), path_(path), sizeMode_(sizeMode), downloader_(downloader) {}
+	HttpImageFileView(http::RequestManager *requestManager, const std::string &path, UI::ImageSizeMode sizeMode = UI::IS_DEFAULT, bool useIconCache = true, UI::LayoutParams *layoutParams = nullptr)
+		: UI::View(layoutParams), path_(path), sizeMode_(sizeMode), requestManager_(requestManager), useIconCache_(useIconCache) {
+
+		if (useIconCache && g_iconCache.MarkPending(path_)) {
+			const char *acceptMime = "image/png, image/jpeg, image/*; q=0.9, */*; q=0.8";
+			requestManager_->StartDownloadWithCallback(path_, Path(), http::ProgressBarMode::DELAYED, [](http::Request &download) {
+				// Can't touch 'this' in this function! Don't use captures!
+				std::string path = download.url();
+				if (download.ResultCode() == 200) {
+					std::string data;
+					download.buffer().TakeAll(&data);
+					if (!data.empty()) {
+						g_iconCache.InsertIcon(path, IconFormat::PNG, std::move(data));
+					} else {
+						g_iconCache.CancelPending(path);
+					}
+				} else {
+					g_iconCache.CancelPending(path);
+				}
+			}, acceptMime);
+		}
+	}
 
 	~HttpImageFileView() {
-		if (download_)
+		if (download_) {
 			download_->Cancel();
+		}
 	}
 
 	void GetContentDimensions(const UIContext &dc, float &w, float &h) const override;
 	void Draw(UIContext &dc) override;
+	std::string DescribeText() const override { return ""; }
 
 	void SetFilename(std::string filename);
 	void SetColor(uint32_t color) { color_ = color; }
@@ -74,14 +101,15 @@ public:
 	const std::string &GetFilename() const { return path_; }
 
 private:
-	void DownloadCompletedCallback(http::Download &download);
+	void DownloadCompletedCallback(http::Request &download);
 
 	bool canFocus_ = false;
-	std::string path_;
+	bool useIconCache_ = false;
+	std::string path_;  // or cache key
 	uint32_t color_ = 0xFFFFFFFF;
 	UI::ImageSizeMode sizeMode_;
-	http::Downloader *downloader_;
-	std::shared_ptr<http::Download> download_;
+	http::RequestManager *requestManager_;
+	std::shared_ptr<http::Request> download_;
 
 	std::string textureData_;
 	std::unique_ptr<ManagedTexture> texture_;
@@ -98,28 +126,39 @@ void HttpImageFileView::GetContentDimensions(const UIContext &dc, float &w, floa
 		break;
 	case UI::IS_DEFAULT:
 	default:
-		if (texture_) {
-			float texw = (float)texture_->Width();
-			float texh = (float)texture_->Height();
-			w = texw;
-			h = texh;
+		if (useIconCache_) {
+			int width, height;
+			if (g_iconCache.GetDimensions(path_, &width, &height)) {
+				w = width;
+				h = height;
+			} else {
+				w = 16;
+				h = 16;
+			}
 		} else {
-			w = 16;
-			h = 16;
+			if (texture_) {
+				float texw = (float)texture_->Width();
+				float texh = (float)texture_->Height();
+				w = texw;
+				h = texh;
+			} else {
+				w = 16;
+				h = 16;
+			}
 		}
 		break;
 	}
 }
 
 void HttpImageFileView::SetFilename(std::string filename) {
-	if (path_ != filename) {
+	if (!useIconCache_ && path_ != filename) {
 		textureFailed_ = false;
 		path_ = filename;
 		texture_.reset(nullptr);
 	}
 }
 
-void HttpImageFileView::DownloadCompletedCallback(http::Download &download) {
+void HttpImageFileView::DownloadCompletedCallback(http::Request &download) {
 	if (download.IsCancelled()) {
 		// We were probably destroyed. Can't touch "this" (heh).
 		return;
@@ -133,17 +172,21 @@ void HttpImageFileView::DownloadCompletedCallback(http::Download &download) {
 
 void HttpImageFileView::Draw(UIContext &dc) {
 	using namespace Draw;
-	if (!texture_ && !textureFailed_ && !path_.empty() && !download_) {
-		download_ = downloader_->StartDownloadWithCallback(path_, "", std::bind(&HttpImageFileView::DownloadCompletedCallback, this, std::placeholders::_1));
-		download_->SetHidden(true);
-	}
 
-	if (!textureData_.empty()) {
-		texture_ = CreateTextureFromFileData(dc.GetDrawContext(), (const uint8_t *)(textureData_.data()), (int)textureData_.size(), DETECT);
-		if (!texture_)
-			textureFailed_ = true;
-		textureData_.clear();
-		download_.reset();
+	if (!useIconCache_) {
+		if (!texture_ && !textureFailed_ && !path_.empty() && !download_) {
+			auto cb = std::bind(&HttpImageFileView::DownloadCompletedCallback, this, std::placeholders::_1);
+			const char *acceptMime = "image/png, image/jpeg, image/*; q=0.9, */*; q=0.8";
+			requestManager_->StartDownloadWithCallback(path_, Path(), http::ProgressBarMode::NONE, cb, acceptMime);
+		}
+
+		if (!textureData_.empty()) {
+			texture_ = CreateTextureFromFileData(dc.GetDrawContext(), (const uint8_t *)(textureData_.data()), (int)textureData_.size(), DETECT, false, "store_icon");
+			if (!texture_)
+				textureFailed_ = true;
+			textureData_.clear();
+			download_.reset();
+		}
 	}
 
 	if (HasFocus()) {
@@ -151,9 +194,16 @@ void HttpImageFileView::Draw(UIContext &dc) {
 	}
 
 	// TODO: involve sizemode
-	if (texture_) {
-		float tw = texture_->Width();
-		float th = texture_->Height();
+	Draw::Texture *texture = nullptr;
+	if (useIconCache_) {
+		texture = g_iconCache.BindIconTexture(&dc, path_);
+	} else {
+		texture = texture_->GetTexture();
+	}
+
+	if (texture) {
+		float tw = texture->Width();
+		float th = texture->Height();
 
 		float x = bounds_.x;
 		float y = bounds_.y;
@@ -171,7 +221,7 @@ void HttpImageFileView::Draw(UIContext &dc) {
 		}
 
 		dc.Flush();
-		dc.GetDrawContext()->BindTexture(0, texture_->GetTexture());
+		dc.GetDrawContext()->BindTexture(0, texture);
 		dc.Draw()->Rect(x, y, w, h, color_);
 		dc.Flush();
 		dc.RebindTexture();
@@ -197,7 +247,7 @@ public:
 	StoreEntry GetEntry() const { return entry_; }
 
 private:
-	const StoreEntry &entry_;
+	const StoreEntry entry_;
 };
 
 // This is a "details" view of a game. Lets you install it.
@@ -222,11 +272,13 @@ private:
 	bool IsGameInstalled() {
 		return g_GameManager.IsGameInstalled(entry_.file);
 	}
+	std::string DownloadURL();
 
 	StoreEntry entry_;
 	UI::Button *installButton_ = nullptr;
 	UI::Button *launchButton_ = nullptr;
 	UI::Button *cancelButton_ = nullptr;
+	UI::TextView *speedView_ = nullptr;
 	bool wasInstalled_ = false;
 };
 
@@ -235,19 +287,27 @@ void ProductView::CreateViews() {
 	Clear();
 
 	if (!entry_.iconURL.empty()) {
-		Add(new HttpImageFileView(&g_DownloadManager, ResolveUrl(storeBaseUrl, entry_.iconURL), IS_FIXED))->SetFixedSize(144, 88);
+		Add(new HttpImageFileView(&g_DownloadManager, ResolveUrl(StoreBaseUrl(), entry_.iconURL), IS_FIXED))->SetFixedSize(144, 88);
 	}
 	Add(new TextView(entry_.name));
 	Add(new TextView(entry_.author));
 
-	I18NCategory *st = GetI18NCategory("Store");
-	I18NCategory *di = GetI18NCategory("Dialog");
+	auto st = GetI18NCategory(I18NCat::STORE);
+	auto di = GetI18NCategory(I18NCat::DIALOG);
 	wasInstalled_ = IsGameInstalled();
+	bool isDownloading = g_GameManager.IsDownloading(DownloadURL());
 	if (!wasInstalled_) {
-		installButton_ = Add(new Button(st->T("Install")));
+		launchButton_ = nullptr;
+		LinearLayout *progressDisplay = new LinearLayout(ORIENT_HORIZONTAL);
+		installButton_ = progressDisplay->Add(new Button(st->T("Install")));
 		installButton_->OnClick.Handle(this, &ProductView::OnInstall);
+
+		speedView_ = progressDisplay->Add(new TextView(""));
+		speedView_->SetVisibility(isDownloading ? V_VISIBLE : V_GONE);
+		Add(progressDisplay);
 	} else {
 		installButton_ = nullptr;
+		speedView_ = nullptr;
 		Add(new TextView(st->T("Already Installed")));
 		Add(new Button(st->T("Uninstall")))->OnClick.Handle(this, &ProductView::OnUninstall);
 		launchButton_ = new Button(st->T("Launch Game"));
@@ -257,16 +317,19 @@ void ProductView::CreateViews() {
 
 	cancelButton_ = Add(new Button(di->T("Cancel")));
 	cancelButton_->OnClick.Handle(this, &ProductView::OnCancel);
-	cancelButton_->SetVisibility(V_GONE);
+	cancelButton_->SetVisibility(isDownloading ? V_VISIBLE : V_GONE);
 
 	// Add star rating, comments etc?
-	Add(new TextView(entry_.description, ALIGN_LEFT | FLAG_WRAP_TEXT, false));
+
+	// Draw each line separately so focusing can scroll.
+	std::vector<std::string> lines;
+	SplitString(entry_.description, '\n', lines);
+	for (auto &line : lines) {
+		Add(new TextView(line, ALIGN_LEFT | FLAG_WRAP_TEXT, false))->SetFocusable(true);
+	}
 
 	float size = entry_.size / (1024.f * 1024.f);
-	char temp[256];
-	sprintf(temp, "%s: %.2f %s", st->T("Size"), size, st->T("MB"));
-
-	Add(new TextView(temp));
+	Add(new TextView(StringFromFormat("%s: %.2f %s", st->T("Size"), size, st->T("MB"))));
 }
 
 void ProductView::Update() {
@@ -276,30 +339,46 @@ void ProductView::Update() {
 	if (installButton_) {
 		installButton_->SetEnabled(g_GameManager.GetState() == GameManagerState::IDLE);
 	}
-	if (cancelButton_ && g_GameManager.GetState() != GameManagerState::DOWNLOADING)
-		cancelButton_->SetVisibility(UI::V_GONE);
+	if (g_GameManager.GetState() == GameManagerState::DOWNLOADING) {
+		if (speedView_) {
+			float speed = g_GameManager.DownloadSpeedKBps();
+			speedView_->SetText(StringFromFormat("%0.1f KB/s", speed));
+		}
+	} else {
+		if (cancelButton_)
+			cancelButton_->SetVisibility(UI::V_GONE);
+		if (speedView_)
+			speedView_->SetVisibility(UI::V_GONE);
+	}
 	if (launchButton_)
 		launchButton_->SetEnabled(g_GameManager.GetState() == GameManagerState::IDLE);
 	View::Update();
 }
 
-UI::EventReturn ProductView::OnInstall(UI::EventParams &e) {
-	std::string zipUrl;
+std::string ProductView::DownloadURL() {
 	if (entry_.downloadURL.empty()) {
-		// Construct the URL, easy to predict from our server
-		zipUrl = storeBaseUrl + "files/" + entry_.file + ".zip";
+		// Construct the URL.
+		return StoreBaseUrl() + "files/" + entry_.file + ".zip";
 	} else {
 		// Use the provided URL, for external hosting.
-		zipUrl = entry_.downloadURL;
+		return entry_.downloadURL;
 	}
+}
+
+UI::EventReturn ProductView::OnInstall(UI::EventParams &e) {
+	std::string fileUrl = DownloadURL();
 	if (installButton_) {
 		installButton_->SetEnabled(false);
 	}
 	if (cancelButton_) {
 		cancelButton_->SetVisibility(UI::V_VISIBLE);
 	}
-	INFO_LOG(SYSTEM, "Triggering install of %s", zipUrl.c_str());
-	g_GameManager.DownloadAndInstall(zipUrl);
+	if (speedView_) {
+		speedView_->SetVisibility(UI::V_VISIBLE);
+		speedView_->SetText("");
+	}
+	INFO_LOG(SYSTEM, "Triggering install of '%s'", fileUrl.c_str());
+	g_GameManager.DownloadAndInstall(fileUrl);
 	return UI::EVENT_DONE;
 }
 
@@ -320,28 +399,23 @@ UI::EventReturn ProductView::OnLaunchClick(UI::EventParams &e) {
 		return UI::EVENT_DONE;
 	}
 
-	std::string pspGame = GetSysDirectory(DIRECTORY_GAME);
-	std::string path = pspGame + entry_.file;
-#ifdef _WIN32
-	path = ReplaceAll(path, "\\", "/");
-#endif
+	Path pspGame = GetSysDirectory(DIRECTORY_GAME);
+	Path path = pspGame / entry_.file;
 	UI::EventParams e2{};
 	e2.v = e.v;
-	e2.s = path;
+	e2.s = path.ToString();
 	// Insta-update - here we know we are already on the right thread.
 	OnClickLaunch.Trigger(e2);
 	return UI::EVENT_DONE;
 }
 
 StoreScreen::StoreScreen() {
-	StoreFilter noFilter;
-	SetFilter(noFilter);
 	lang_ = g_Config.sLanguageIni;
 	loading_ = true;
 
-	std::string indexPath = storeBaseUrl + "index.json";
-
-	listing_ = g_DownloadManager.StartDownload(indexPath, "");
+	std::string indexPath = StoreBaseUrl() + "index.json";
+	const char *acceptMime = "application/json, */*; q=0.8";
+	listing_ = g_DownloadManager.StartDownload(indexPath, Path(), http::ProgressBarMode::DELAYED, acceptMime);
 }
 
 StoreScreen::~StoreScreen() {
@@ -355,6 +429,7 @@ void StoreScreen::update() {
 	g_DownloadManager.Update();
 
 	if (listing_.get() != 0 && listing_->Done()) {
+		resultCode_ = listing_->ResultCode();
 		if (listing_->ResultCode() == 200) {
 			std::string listingJson;
 			listing_->buffer().TakeAll(&listingJson);
@@ -366,7 +441,7 @@ void StoreScreen::update() {
 			RecreateViews();
 		} else {
 			// Failed to contact store. Don't do anything.
-			ELOG("Download failed : error code %d", listing_->ResultCode());
+			ERROR_LOG(IO, "Download failed : error code %d", resultCode_);
 			connectionError_ = true;
 			loading_ = false;
 			RecreateViews();
@@ -375,26 +450,13 @@ void StoreScreen::update() {
 		// Forget the listing.
 		listing_.reset();
 	}
-
-	const char *storeName = "PPSSPP Homebrew Store";
-	switch (g_GameManager.GetState()) {
-	case GameManagerState::DOWNLOADING:
-		titleText_->SetText(std::string(storeName) + " - downloading");
-		break;
-	case GameManagerState::INSTALLING:
-		titleText_->SetText(std::string(storeName) + " - installing");
-		break;
-	default:
-		titleText_->SetText(storeName);
-		break;
-	}
 }
 
 void StoreScreen::ParseListing(std::string json) {
 	using namespace json;
 	JsonReader reader(json.c_str(), json.size());
 	if (!reader.ok() || !reader.root()) {
-		ELOG("Error parsing JSON from store");
+		ERROR_LOG(IO, "Error parsing JSON from store");
 		connectionError_ = true;
 		RecreateViews();
 		return;
@@ -413,7 +475,7 @@ void StoreScreen::ParseListing(std::string json) {
 			e.size = game.getInt("size");
 			e.downloadURL = game.getString("download-url", "");
 			e.iconURL = game.getString("icon-url", "");
-			e.hidden = game.getBool("hidden", false);
+			e.hidden = false;  // NOTE: Handling of the "hidden" flag is broken in old versions of PPSSPP. Do not use.
 			const char *file = game.getString("file", nullptr);
 			if (!file)
 				continue;
@@ -428,13 +490,14 @@ void StoreScreen::CreateViews() {
 
 	root_ = new LinearLayout(ORIENT_VERTICAL);
 	
-	I18NCategory *di = GetI18NCategory("Dialog");
-	I18NCategory *st = GetI18NCategory("Store");
+	auto di = GetI18NCategory(I18NCat::DIALOG);
+	auto st = GetI18NCategory(I18NCat::STORE);
+	auto mm = GetI18NCategory(I18NCat::MAINMENU);
 
 	// Top bar
 	LinearLayout *topBar = root_->Add(new LinearLayout(ORIENT_HORIZONTAL));
 	topBar->Add(new Button(di->T("Back")))->OnClick.Handle<UIScreen>(this, &UIScreen::OnBack);
-	titleText_ = new TextView("PPSSPP Homebrew Store");
+	titleText_ = new TextView(mm->T("PPSSPP Homebrew Store"));
 	topBar->Add(titleText_);
 	UI::Drawable solid(0xFFbd9939);
 	topBar->SetBG(solid);
@@ -442,8 +505,11 @@ void StoreScreen::CreateViews() {
 	LinearLayout *content;
 	if (connectionError_ || loading_) {
 		content = new LinearLayout(ORIENT_VERTICAL, new LinearLayoutParams(FILL_PARENT, FILL_PARENT, 1.0f));
-		content->Add(new TextView(loading_ ? st->T("Loading...") : st->T("Connection Error")));
-		content->Add(new Button(di->T("Retry")))->OnClick.Handle(this, &StoreScreen::OnRetry);
+		content->Add(new TextView(loading_ ? std::string(st->T("Loading...")) : StringFromFormat("%s: %d", st->T("Connection Error"), resultCode_)));
+		if (!loading_) {
+			content->Add(new Button(di->T("Retry")))->OnClick.Handle(this, &StoreScreen::OnRetry);
+
+		}
 		content->Add(new Button(di->T("Back")))->OnClick.Handle<UIScreen>(this, &UIScreen::OnBack);
 
 		scrollItemView_ = nullptr;
@@ -453,11 +519,10 @@ void StoreScreen::CreateViews() {
 		ScrollView *leftScroll = new ScrollView(ORIENT_VERTICAL, new LinearLayoutParams(WRAP_CONTENT, FILL_PARENT, 0.4f));
 		leftScroll->SetTag("StoreMainList");
 		content->Add(leftScroll);
-		scrollItemView_ = new LinearLayout(ORIENT_VERTICAL, new LayoutParams(FILL_PARENT, WRAP_CONTENT));
+		scrollItemView_ = new LinearLayoutList(ORIENT_VERTICAL, new LayoutParams(FILL_PARENT, WRAP_CONTENT));
 		leftScroll->Add(scrollItemView_);
 
-		std::vector<StoreEntry> entries = FilterEntries();
-		for (size_t i = 0; i < entries.size(); i++) {
+		for (size_t i = 0; i < entries_.size(); i++) {
 			scrollItemView_->Add(new ProductItemView(entries_[i]))->OnClick.Handle(this, &StoreScreen::OnGameSelected);
 		}
 
@@ -474,22 +539,12 @@ void StoreScreen::CreateViews() {
 
 			selectedItem->Press();
 		} else {
-			lastSelectedName_ = "";
+			lastSelectedName_.clear();
 		}
 	}
 	root_->Add(content);
 }
 
-std::vector<StoreEntry> StoreScreen::FilterEntries() {
-	std::vector<StoreEntry> filtered;
-	for (size_t i = 0; i < entries_.size(); i++) {
-		// TODO: Actually filter by category etc.
-		if (!entries_[i].hidden)
-			filtered.push_back(entries_[i]);
-	}
-	return filtered;
-}
-
 ProductItemView *StoreScreen::GetSelectedItem() {
 	for (int i = 0; i < scrollItemView_->GetNumSubviews(); ++i) {
 		ProductItemView *item = static_cast<ProductItemView *>(scrollItemView_->GetViewByIndex(i));
@@ -519,28 +574,15 @@ UI::EventReturn StoreScreen::OnGameSelected(UI::EventParams &e) {
 
 UI::EventReturn StoreScreen::OnGameLaunch(UI::EventParams &e) {
 	std::string path = e.s;
-	screenManager()->switchScreen(new EmuScreen(path));
+	screenManager()->switchScreen(new EmuScreen(Path(path)));
 	return UI::EVENT_DONE;
 }
 
-void StoreScreen::SetFilter(const StoreFilter &filter) {
-	filter_ = filter;
-	RecreateViews();
-}
-
 UI::EventReturn StoreScreen::OnRetry(UI::EventParams &e) {
-	SetFilter(filter_);
+	RecreateViews();
 	return UI::EVENT_DONE;
 }
 
-std::string StoreScreen::GetStoreJsonURL(std::string storePath) const {
-	std::string path = storeBaseUrl + storePath;
-	if (*path.rbegin() != '/')
-		path += '/';
-	path += "index.json";
-	return path;
-}
-
 std::string StoreScreen::GetTranslatedString(const json::JsonGet json, std::string key, const char *fallback) const {
 	json::JsonGet dict = json.getDict("en_US");
 	if (dict && json.hasChild(lang_.c_str(), JSON_OBJECT)) {
