@@ -1,12 +1,10 @@
-// NOTE: Apologies for the quality of this code, this is really from pre-opensource Dolphin - that is, 2003.
-
 #pragma once
 #include "Windows/W32Util/DialogManager.h"
 
 #include "Core/MemMap.h"
 
 #include "Core/Debugger/DebugInterface.h"
-
+#include "CtrlMemView.h"
 #include "Common/CommonWindows.h"
 
 class CMemoryDlg : public Dialog
@@ -14,11 +12,18 @@ class CMemoryDlg : public Dialog
 private:
 	DebugInterface *cpu;
 	static RECT slRect;
+	RECT winRect, srRect;
+	CtrlMemView *memView;
+	HWND memViewHdl, symListHdl, editWnd, searchBoxHdl, srcListHdl;
+	HWND layerDropdown_;
+	HWND status_;
+	BOOL DlgProc(UINT message, WPARAM wParam, LPARAM lParam) override;
 
-	BOOL DlgProc(UINT message, WPARAM wParam, LPARAM lParam);
 public:
 	int index; //helper 
 
+	void searchBoxRedraw(std::vector<u32> results);
+
 	// constructor
 	CMemoryDlg(HINSTANCE _hInstance, HWND _hParent, DebugInterface *_cpu);
 	
@@ -26,10 +31,13 @@ public:
 	~CMemoryDlg(void);
 	
 	void Goto(u32 addr);
-	void Update(void);	
+	void Update(void) override;
 	void NotifyMapLoaded();
 
 	void Size(void);
+
+private:
+	bool mapLoadPending_ = false;
 };
 
 
